import argparse

import numpy as np


def loguniform(low=-1, high=3, size=None, base=10):
    return np.power(base, np.random.uniform(low, high, size))


parser = argparse.ArgumentParser("Generate some samples!")
parser.add_argument("-n", help="number of samples", default=100, type=int)
parser.add_argument("-outfile", help="name of output .npy file", default="samples")

args = parser.parse_args()

N_SAMPLES = args.n
REYNOLD_RANGE = [1, 100]
LIDSPEED_RANGE = [0.1, 100]
BASE = 10

x = np.empty((N_SAMPLES, 2))
x[:, 0] = np.random.uniform(LIDSPEED_RANGE[0], LIDSPEED_RANGE[1], size=N_SAMPLES)
vi_low = np.log10(x[:, 0] / REYNOLD_RANGE[1])
vi_high = np.log10(x[:, 0] / REYNOLD_RANGE[0])
x[:, 1] = [loguniform(low=vi_low[i], high=vi_high[i], base=BASE) for i in range(N_SAMPLES)]

np.save(args.outfile, x)



import argparse
import datetime
import glob
import os
import re
import subprocess
import sys


SAMPLES = [1, 10, 100, 1000, 10000, 100000]

# argument parsing
parser = argparse.ArgumentParser(description="Make some samples (names of people).")
parser.add_argument("output_path", type=str, help="path to original output")
parser.add_argument("split_data_path", type=str, help="path to split data")
parser.add_argument("c", type=int, help="concurrency")
args = parser.parse_args()

filled_c = str(args.c).zfill(3)

logpattern = os.path.join(args.split_data_path, "split_log_*")
errpattern = os.path.join(args.output_path, "*.err")
args.logfile = glob.glob(logpattern)
args.errfile = glob.glob(errpattern)

if len(args.logfile) == 0:
    print(f"{logpattern} returned no glob matches!")
    sys.exit()
if len(args.errfile) == 0:
    print(f"{errpattern} returned no glob matches!")
    sys.exit()

logmap = {}
i = 1
for log in args.logfile:
    if i == 1000000:
        break
    logmap[i] = log
    i *= 10


def single_task_times():
    for k, v in logmap.items():
        task_durations = []
        try:
            pre_lines = subprocess.check_output(f'grep " succeeded in " {v}', shell=True).decode("ascii")

            pre_list = pre_lines.strip().split("\n")

            for line in pre_list:
                matches = re.search(r"\d+.\d+s:", line)
                if matches:
                    match = matches.group(0)
                    match = float(match.strip("s:"))
                    task_durations.append(match)
        except Exception:
            print(f"c{filled_c}_s{k} task times : ERROR")
            continue

        print(f"c{filled_c}_s{k} task times : " + str(task_durations))


def merlin_run_time():
    total = 0
    for err in args.errfile:
        pre_line = subprocess.check_output(f'grep "real" {err}', shell=True).decode("ascii")
        pre_line = pre_line.strip()
        matches = re.search(r"\d\.\d\d\d", pre_line)
        match = matches[0]
        result = float(match)
        total += result
    try:
        print(f"c{filled_c} merlin run : " + str(result))
    except Exception:
        result = None
        print(f"c{filled_c} merlin run : ERROR -- result={result}, args.errfile={args.errfile}")


def start_verify_time():
    for k, v in logmap.items():
        all_timestamps = []
        try:
            pre_line = subprocess.check_output(f'grep -m2 "verify" {v} | tail -n1', shell=True).decode("ascii")
            pre_line = pre_line.strip()
            matches = re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", pre_line)
            match = matches[0]
            element = datetime.datetime.strptime(match, "%Y-%m-%d %H:%M:%S,%f")
            timestamp = datetime.datetime.timestamp(element)
            all_timestamps.append(timestamp)
        except Exception:
            print(f"c{filled_c}_s{k} start verify : ERROR")
            continue
        try:
            print(f"c{filled_c}_s{k} start verify : " + str(all_timestamps[0]))
        except Exception:
            print(f"c{filled_c}_s{k} start verify : ERROR")


def start_run_workers_time():
    for k, v in logmap.items():
        all_timestamps = []
        try:
            pre_line = subprocess.check_output(f'grep -m1 "" {v}', shell=True).decode("ascii")
            pre_line = pre_line.strip()
            matches = re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", pre_line)
            match = matches[0]
            element = datetime.datetime.strptime(match, "%Y-%m-%d %H:%M:%S,%f")
            timestamp = datetime.datetime.timestamp(element)
            all_timestamps.append(timestamp)
        except Exception:
            continue
        earliest = min(all_timestamps)
        print(f"c{filled_c}_s{k} start run-workers : " + str(earliest))


def start_sample1_time():
    for k, v in logmap.items():
        all_timestamps = []
        try:
            pre_line = subprocess.check_output(f"grep -m1 \"Executing step 'null_step'\" {v}", shell=True).decode("ascii")
            pre_line = pre_line.strip()
            matches = re.search(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d", pre_line)
            match = matches[0]
            element = datetime.datetime.strptime(match, "%Y-%m-%d %H:%M:%S,%f")
            timestamp = datetime.datetime.timestamp(element)
            all_timestamps.append(timestamp)
        except Exception:
            print(f"c{filled_c}_s{k} start samp1 : ERROR")
            continue
        earliest = min(all_timestamps)
        print(f"c{filled_c}_s{k} start samp1 : " + str(earliest))


def main():
    single_task_times()
    merlin_run_time()
    start_verify_time()
    start_run_workers_time()
    start_sample1_time()


if __name__ == "__main__":
    main()



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################



#!/usr/bin/env python3
"""This module will collect information on flux jobs from the live kvs
store and output times for each phase.

old:
create: The time flux registerd the job
starting: The time the job was created
running: The time the job was running
completing: The time the job started its' completion pahse.
complete: The time the job was complete
walltime: ? Seems to be 0.

new:
init:
starting:
shell.init:
shell.start:
complete:
cleanup.start:
cleanup.finish:
done:
"""
import json
import os
import subprocess
from typing import IO, Dict, Union

import flux
from flux import kvs


f = flux.Flux()

fs = "FLUX_START_SECONDS"
if fs in os.environ:
    print("flux start: {0}".format(os.environ[fs]))

try:
    kvs.get(f, "lwj")

    for d in kvs.walk("lwj", flux_handle=f):
        # print(type(d))
        fdir = "lwj.{0}".format(d[0])

        qcreate = "{0}.create-time".format(fdir)
        create_time = kvs.get(f, qcreate)

        qstart = "{0}.starting-time".format(fdir)
        start_time = kvs.get(f, qstart)

        qrun = "{0}.running-time".format(fdir)
        start_time = kvs.get(f, qrun)

        qcomplete = "{0}.complete-time".format(fdir)
        complete_time = kvs.get(f, qcomplete)

        qcompleting = "{0}.completing-time".format(fdir)
        completing_time = kvs.get(f, qcompleting)

        qwall = "{0}.walltime".format(fdir)
        wall_time = kvs.get(f, qwall)

        print(
            f"Job {d[0]}: create: {create_time} start {start_time} run {start_time} completing {completing_time} complete {complete_time} wall {wall_time}"
        )

except KeyError:
    top_dir = "job"

    def get_data_dict(key: str) -> Dict:
        kwargs: Dict[str, Union[str, bool, os.Environ]] = {
            "env": os.environ,
            "shell": True,
            "universal_newlines": True,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.PIPE,
        }
        flux_com: str = f"flux kvs get {key}"
        p: subprocess.Popen = subprocess.Popen(flux_com, **kwargs)
        stdout: IO[str]
        stderr: IO[str]
        stdout, stderr = p.communicate()

        data: Dict = {}
        line: str
        for line in stdout.split("/n"):
            token: str
            for token in line.strip().split():
                if "timestamp" in token:
                    jstring: str = token.replace("'", '"')
                    d: Dict = json.loads(jstring)
                    data[d["name"]] = d["timestamp"]

        return data

    for d in kvs.walk(top_dir, flux_handle=f):
        if "exec" in d[0]:
            for e in d[2]:
                key = ".".join([top_dir, d[0], e])

                # This is currently not working gives
                # json.decoder.JSONDecodeError
                # data = kvs.get(f, key)

                data = get_data_dict(key)

                print(
                    f"Job {d[0]}: init: {data['init']} start {data['shell.start']} complete {data['complete']} done {data['done']} "
                )


# vi: ts=4 sw=4 expandtab



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for querying active Merlin task server workers.

This module defines the `QueryWorkersCommand` class, which implements the
`query-workers` subcommand for the Merlin CLI. The command allows users to
inspect the state of connected workers on a task server (e.g., Celery),
optionally filtering by queues or worker names.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentParser, Namespace

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.router import query_workers
from merlin.spec.specification import MerlinSpec
from merlin.utils import verify_filepath


LOG = logging.getLogger("merlin")


class QueryWorkersCommand(CommandEntryPoint):
    """
    Handles `query-workers` CLI command for querying information about Merlin workers.

    Methods:
        add_parser: Adds the `query-workers` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `query-workers` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `query-workers` command parser will be added.
        """
        query: ArgumentParser = subparsers.add_parser("query-workers", help="List connected task server workers.")
        query.set_defaults(func=self.process_command)
        query.add_argument(
            "--task_server",
            type=str,
            default="celery",
            help="Task server type from which to query workers.\
                                Default: %(default)s",
        )
        query.add_argument(
            "--spec",
            type=str,
            default=None,
            help="Path to a Merlin YAML spec file from which to read worker names to query.",
        )
        query.add_argument("--queues", type=str, default=None, nargs="+", help="Specific queues to query workers from.")
        query.add_argument(
            "--workers",
            type=str,
            action="store",
            nargs="+",
            default=None,
            help="Regex match for specific workers to query.",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for finding all workers.

        This function retrieves and queries the names of any active workers.
        If the `--spec` argument is included, only query the workers defined in the spec file.

        Args:
            args: Parsed command-line arguments, which may include:\n
                - `spec`: Path to the specification file.
                - `task_server`: Address of the task server to query.
                - `queues`: List of queue names to filter workers.
                - `workers`: List of specific worker names to query.
        """
        print(banner_small)

        # Get the workers from the spec file if --spec provided
        worker_names = []
        if args.spec:
            spec_path = verify_filepath(args.spec)
            spec = MerlinSpec.load_specification(spec_path)
            worker_names = spec.get_worker_names()
            for worker_name in worker_names:
                if "$" in worker_name:
                    LOG.warning(f"Worker '{worker_name}' is unexpanded. Target provenance spec instead?")
            LOG.debug(f"Searching for the following workers to stop based on the spec {args.spec}: {worker_names}")

        query_workers(args.task_server, worker_names, args.queues, args.workers)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module provides the `Monitor` class, which is responsible for monitoring the progress of
Merlin workflows. It ensures that workers are running, tasks are being processed, and workflows
are restarted if needed to prevent hanging. The `Monitor` class uses worker and task monitors
to manage the health and progress of workflows.

The module interacts with the Merlin database to retrieve study and run information and
uses the `monitor_factory` to create monitors for task and worker systems (e.g., Celery).

Exceptions such as Redis timeouts, Kombu operational errors, and other runtime issues are
handled gracefully to ensure that monitoring continues without interruption.
"""

import logging
import subprocess
import time
import traceback

from kombu.exceptions import OperationalError
from redis.exceptions import TimeoutError as RedisTimeoutError

from merlin.db_scripts.entities.run_entity import RunEntity
from merlin.db_scripts.merlin_db import MerlinDatabase
from merlin.exceptions import RestartException
from merlin.monitor.monitor_factory import monitor_factory
from merlin.monitor.task_server_monitor import TaskServerMonitor
from merlin.spec.specification import MerlinSpec
from merlin.utils import verify_dirpath


LOG = logging.getLogger(__name__)


class Monitor:
    """
    The `Monitor` class is responsible for monitoring the progress of Merlin workflows. It ensures
    that workers are running, tasks are being processed, and workflows are restarted if necessary
    to prevent hanging. As a side-effect of the monitor, the users allocation will remain alive for
    however long the monitor lives. The class interacts with the Merlin database to retrieve study
    and run information and uses a task server monitor to help manage workflow health.

    Attributes:
        spec (MerlinSpec): The Merlin specification that defines the workflow.
        sleep (int): The interval (in seconds) between monitoring checks.
        no_restart (bool): If True, the monitor will not try to restart the workflow.
        task_server_monitor (TaskServerMonitor): A monitor for interacting with whichever task server
            that the user is utilizing.
        merlin_db (MerlinDatabase): Interface for accessing and querying the Merlin database.

    Methods:
        monitor_all_runs: Monitors all runs of the current study until they are complete.
        monitor_single_run: Monitors a single run of a study until it completes.
        restart_workflow: Restart a run of a workflow.
    """

    def __init__(self, spec: MerlinSpec, sleep: int, task_server: str, no_restart: bool):
        """
        Initializes the `Monitor` instance with the given Merlin specification, sleep interval,
        and task server type. The task server monitor is created using the
        [`monitor_factory`][monitor.monitor_factory.MonitorFactory].

        Args:
            spec (MerlinSpec): The Merlin specification that defines the workflow.
            sleep (int): The interval (in seconds) between monitoring checks.
            task_server (str): The type of task server being used (e.g., "celery").
            no_restart (bool): If True, the monitor will not try to restart the workflow.
        """
        self.spec: MerlinSpec = spec
        self.sleep: int = sleep
        self.no_restart: bool = no_restart
        self.task_server_monitor: TaskServerMonitor = monitor_factory.get_monitor(task_server)
        self.merlin_db = MerlinDatabase()

    def monitor_all_runs(self):
        """
        Monitors all runs of the current study until they are complete. For each run, it checks
        if the run is already complete. If not, it monitors the run until it finishes. This
        method ensures that all runs in the study are processed. This is necessary to be able to
        monitor iterative workflows.

        The method retrieves all runs from the database and iterates through them sequentially.
        If a run is incomplete, it calls [`monitor_single_run`][monitor.monitor.Monitor.monitor_single_run]
        to monitor it until completion.
        """
        study_entity = self.merlin_db.get("study", self.spec.name)

        index = 0
        while True:
            # Always refresh the list at the start of the loop; there could be new runs (think iterative studies)
            all_runs = [self.merlin_db.get("run", run_id) for run_id in study_entity.get_runs()]
            if index >= len(all_runs):  # Break if there are no more runs to process
                break

            run = all_runs[index]
            run_workspace = run.get_workspace()
            LOG.info(f"Monitor: Checking if run with workspace '{run_workspace}' has completed...")

            if run.run_complete:
                LOG.info(
                    f"Monitor: Determined that run with workspace '{run_workspace}' has already completed. "
                    "Moving on to the next run."
                )
                index += 1
                continue

            LOG.info(f"Monitor: Run with workspace '{run_workspace}' has not yet completed.")

            # Monitor the run until it completes
            self.monitor_single_run(run)

            index += 1

    def _check_task_activity(self, run: RunEntity) -> bool:
        """
        Checks whether there is active task activity for the given run.

        This method first checks if there are any tasks in the task server's queues. If not,
        it then checks whether any workers are currently processing tasks. If either of these
        conditions is true, the method considers the workflow to be active and returns True.

        Args:
            run (RunEntity): The run entity representing the workflow run to check for activity.

        Returns:
            True if tasks are in the queues or being processed by workers, False otherwise.
        """
        # Check if any tasks are currently in the queues
        if self.task_server_monitor.check_tasks(run):
            LOG.info("Monitor: Found tasks in queues, keeping allocation alive.")
            return True

        # If no tasks are in the queues, check if workers are processing tasks
        if self.task_server_monitor.check_workers_processing(run.get_queues()):
            LOG.info("Monitor: Found workers processing tasks, keeping allocation alive.")
            return True

        return False

    def _handle_transient_exception(self, exc: Exception):
        """
        Handles transient exceptions that may occur during monitoring.

        This method logs the exception type, message, and full traceback, then
        sleeps for the configured interval before retrying. It is designed to
        gracefully handle recoverable errors such as Redis timeouts or broker issues.

        Args:
            exc (Exception): The exception instance that was caught during execution.
        """
        LOG.warning(f"{exc.__class__.__name__} occurred:\n{exc}")
        LOG.warning(f"Full traceback:\n{traceback.format_exc()}")
        time.sleep(self.sleep)

    def monitor_single_run(self, run: RunEntity):
        """
        Monitors a single run of a study until it completes to ensure that the allocation stays alive
        and workflows are restarted if necessary.

        Args:
            run: A [`RunEntity`][db_scripts.entities.run_entity.RunEntity] instance representing
                the run that's going to be monitored.
        """
        run_workspace = run.get_workspace()
        run_complete = run.run_complete  # Saving this to a variable as it queries the db each time it's called

        LOG.info(f"Monitor: Monitoring run with workspace '{run_workspace}'...")

        # Wait for workers to spin up before checking on tasks
        worker_names = [self.merlin_db.get("logical_worker", worker_id=wid).get_name() for wid in run.get_workers()]
        LOG.info(f"Monitor: Waiting for the following workers to start: {worker_names}...")
        self.task_server_monitor.wait_for_workers(worker_names, self.sleep)
        LOG.info("Monitor: Workers have started.")

        while not run_complete:
            try:
                # Run worker health check (checks for dead workers and restarts them if necessary)
                self.task_server_monitor.run_worker_health_check(run.get_workers())

                # Check if any tasks are currently in the queues or if workers are processing tasks
                active_tasks = self._check_task_activity(run)

                run_complete = run.run_complete  # Re-query db for this value

                # If no tasks are in the queues or being processed by workers and the run is not complete, we have a hanging
                # workflow so restart it
                if not active_tasks and not run_complete:
                    if self.no_restart:
                        LOG.warning(
                            f"Monitor: Determined restart was required for '{run_workspace}' but auto-restart is disabled."
                        )
                    else:
                        self.restart_workflow(run)

                if not run_complete:
                    time.sleep(self.sleep)
            # The below exceptions do not modify the `run_complete` value so the loop should retry
            except (RedisTimeoutError, OperationalError, TimeoutError) as exc:
                self._handle_transient_exception(exc)

        LOG.info(f"Monitor: Run with workspace '{run_workspace}' has completed.")

    def restart_workflow(self, run: RunEntity):
        """
        Restart a run of a workflow.

        Args:
            run: A [`RunEntity`][db_scripts.entities.run_entity.RunEntity] instance representing
                the run that's going to be restarted.

        Raises:
            RestartException: If the workflow restart process fails.
        """
        try:
            run_workspace = verify_dirpath(run.get_workspace())
            LOG.info(f"Monitor: Restarting workflow for run with workspace '{run_workspace}'...")
            restart_proc = subprocess.run(f"merlin restart {run_workspace}", shell=True, capture_output=True, text=True)
            if restart_proc.returncode != 0:
                LOG.error(f"Monitor: Failed to restart workflow: {restart_proc.stderr}")
                raise RestartException(f"Restart process failed with error: {restart_proc.stderr}")
            LOG.info(f"Monitor: Workflow restarted successfully: {restart_proc.stdout}")
        except ValueError:
            LOG.warning(
                f"Monitor: Run with workspace '{run.get_workspace()}' was not found. Ignoring the restart of this workspace."
            )



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module stores constants representing file paths that will be needed for
Merlin's configuration.
"""

import os


APP_FILENAME: str = "app.yaml"
USER_HOME: str = os.path.expanduser("~")
MERLIN_HOME: str = os.path.join(USER_HOME, ".merlin")
CONFIG_PATH_FILE: str = os.path.join(MERLIN_HOME, "config_path.txt")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Used to store the application configuration.

The `config` package provides functionality for managing and configuring various aspects
of the Merlin application, including broker settings, results backends, Celery configurations,
and application-level settings. It serves as the central hub for loading, processing, and
utilizing configuration data defined in the `app.yaml` file and other related resources.

Modules:
    broker.py: Manages broker configurations and connection strings for messaging systems.
    celeryconfig.py: Contains default Celery configuration settings for Merlin.
    configfile.py: Handles the loading and processing of application configuration files
        and SSL-related settings.
    results_backend.py: Configures connection strings and SSL settings for results backends.
    utils.py: Provides utilities for broker priority handling and validation.
"""
from copy import copy
from types import SimpleNamespace
from typing import Dict, List, Optional

from merlin.utils import nested_dict_to_namespaces


# Pylint complains that there's too few methods here but this class might
# be useful if we ever need to do extra stuff with the configuration so we'll
# ignore it for now
class Config:  # pylint: disable=R0903
    """
    The Config class, meant to store all Merlin config settings in one place.
    Regardless of the config data loading method, this class is meant to
    standardize config data retrieval throughout all parts of Merlin.

    Attributes:
        celery (Optional[SimpleNamespace]): A namespace containing Celery configuration settings.
        broker (Optional[SimpleNamespace]): A namespace containing broker configuration settings.
        results_backend (Optional[SimpleNamespace]): A namespace containing results backend configuration settings.

    Methods:
        __copy__: Creates a shallow copy of the Config instance.
        __str__: Returns a formatted string representation of the Config instance.
        load_app_into_namespaces: Converts the provided configuration dictionary into namespaces
            and assigns them to the Config instance's attributes.
    """

    def __init__(self, app_dict: Dict):
        """
        Initializes the Config instance with configuration data from a dictionary.

        Args:
            app_dict: A dictionary containing configuration data for the application.
                The dictionary may include keys such as "celery", "broker", and "results_backend",
                each of which is converted into a `SimpleNamespace` and assigned to the corresponding
                attribute of the Config instance.
        """
        # I think this ends up a SimpleNamespace from load_app_into_namespaces, but it seems like it should be typed as
        # the app var in celery.py, as celery.app.base.Celery
        self.celery: Optional[SimpleNamespace]
        self.broker: Optional[SimpleNamespace]
        self.results_backend: Optional[SimpleNamespace]
        self.load_app_into_namespaces(app_dict)

    def __copy__(self) -> "Config":
        """
        Creates a shallow copy of the Config instance.

        Returns:
            A new Config instance with copied `celery`, `broker`, and `results_backend` attributes.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        copied_attrs = {
            "celery": copy(self.__dict__["celery"]),
            "broker": copy(self.__dict__["broker"]),
            "results_backend": copy(self.__dict__["results_backend"]),
        }
        result.__dict__.update(copied_attrs)
        return result

    def __str__(self) -> str:
        """
        Returns a formatted string representation of the Config instance.

        Returns:
            str: A string containing the values of the `celery`, `broker`, and `results_backend` attributes.
        """
        formatted_str = "config:"
        attrs = {"celery": self.celery, "broker": self.broker, "results_backend": self.results_backend}
        for name, attr in attrs.items():
            if attr is not None:
                items = (f"    {k}: {v!r}" for k, v in attr.__dict__.items())
                joined_items = "\n".join(items)
                formatted_str += f"\n  {name}:\n{joined_items}"
            else:
                formatted_str += f"\n  {name}:\n    None"
        return formatted_str

    def load_app_into_namespaces(self, app_dict: Dict):
        """
        Converts the provided application dictionary into namespaces and assigns them
        to the Config instance's attributes.

        Args:
            app_dict: A dictionary containing configuration data for the application.
        """
        fields: List[str] = ["celery", "broker", "results_backend"]
        for field in fields:
            try:
                setattr(self, field, nested_dict_to_namespaces(app_dict[field]))
            except KeyError:
                # The keywords are optional
                pass



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for executing Merlin or Maestro workflows.

This module defines the `RunCommand` class, which handles the `run` subcommand in the
Merlin CLI. The command initializes and runs a workflow based on a specified YAML
study specification file. It supports overriding variables, supplying samples files,
dry-run execution, and parallel parameter generation options.
"""

# pylint: disable=duplicate-code

from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from typing import Optional

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.cli.utils import parse_override_vars
from merlin.config.configfile import initialize_config
from merlin.db_scripts.merlin_db import MerlinDatabase
from merlin.router import run_task_server
from merlin.study.study import MerlinStudy
from merlin.utils import ARRAY_FILE_FORMATS, verify_filepath


class RunCommand(CommandEntryPoint):
    """
    Handles `run` CLI command for running a workflow (sending tasks to the queues on the broker).

    Methods:
        add_parser: Adds the `run` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `run` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `run` command parser will be added.
        """
        run: ArgumentParser = subparsers.add_parser(
            "run",
            help="Run a workflow using a Merlin or Maestro YAML study " "specification.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        run.set_defaults(func=self.process_command)
        run.add_argument("specification", type=str, help="Path to a Merlin or Maestro YAML file")
        run.add_argument(
            "--local",
            action="store_const",
            dest="run_mode",
            const="local",
            default="distributed",
            help="Run locally instead of distributed",
        )
        run.add_argument(
            "--vars",
            action="store",
            dest="variables",
            type=str,
            nargs="+",
            default=None,
            help="Specify desired Merlin variable values to override those found in the specification. Space-delimited. "
            "Example: '--vars LEARN=path/to/new_learn.py EPOCHS=3'",
        )
        # TODO add all supported formats to doc string  # pylint: disable=fixme
        run.add_argument(
            "--samplesfile",
            action="store",
            dest="samples_file",
            type=str,
            default=None,
            help=f"Specify file containing samples. Valid choices: {ARRAY_FILE_FORMATS}",
        )
        run.add_argument(
            "--dry",
            action="store_true",
            dest="dry",
            default=False,
            help="Flag to dry-run a workflow, which sets up the workspace but does not launch tasks.",
        )
        run.add_argument(
            "--no-errors",
            action="store_true",
            dest="no_errors",
            default=False,
            help="Flag to ignore some flux errors for testing (often used with --dry --local).",
        )
        run.add_argument(
            "--pgen",
            action="store",
            dest="pgen_file",
            type=str,
            default=None,
            help="Provide a pgen file to override global.parameters.",
        )
        run.add_argument(
            "--pargs",
            type=str,
            action="append",
            default=[],
            help="A string that represents a single argument to pass "
            "a custom parameter generation function. Reuse '--parg' "
            "to pass multiple arguments. [Use with '--pgen']",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for running a study.

        This function initializes and runs a study using the specified parameters.
        It handles file verification, variable parsing, and checks for required
        arguments related to the study configuration and execution.

        Args:
            args: Parsed CLI arguments containing:\n
                - `specification`: Path to the specification file for the study.
                - `variables`: Optional variable overrides for the study.
                - `samples_file`: Optional path to a samples file.
                - `dry`: If True, runs the study in dry-run mode (without actual execution).
                - `no_errors`: If True, suppresses error reporting.
                - `pgen_file`: Optional path to the pgen file, required if `pargs` is specified.
                - `pargs`: Additional arguments for parallel processing.

        Raises:
            ValueError:
                If the `pargs` parameter is used without specifying a `pgen_file`.
        """
        print(banner_small)
        filepath: str = verify_filepath(args.specification)
        variables_dict: str = parse_override_vars(args.variables)
        samples_file: Optional[str] = None
        if args.samples_file:
            samples_file = verify_filepath(args.samples_file)

        # pgen checks
        if args.pargs and not args.pgen_file:
            raise ValueError("Cannot use the 'pargs' parameter without specifying a 'pgen'!")
        if args.pgen_file:
            verify_filepath(args.pgen_file)

        study: MerlinStudy = MerlinStudy(
            filepath,
            override_vars=variables_dict,
            samples_file=samples_file,
            dry_run=args.dry,
            no_errors=args.no_errors,
            pgen_file=args.pgen_file,
            pargs=args.pargs,
        )

        if args.run_mode == "local":
            initialize_config(local_mode=True)

        # Initialize the database
        merlin_db = MerlinDatabase()

        # Create a run entry
        run_entity = merlin_db.create(
            "run",
            study_name=study.expanded_spec.name,
            workspace=study.workspace,
            queues=study.expanded_spec.get_queue_list(["all"]),
        )

        # Create logical worker entries
        step_queue_map = study.expanded_spec.get_task_queues()
        for worker, steps in study.expanded_spec.get_worker_step_map().items():
            worker_queues = {step_queue_map[step] for step in steps}
            logical_worker_entity = merlin_db.create("logical_worker", worker, worker_queues)

            # Add the run id to the worker entry and the worker id to the run entry
            logical_worker_entity.add_run(run_entity.get_id())
            run_entity.add_worker(logical_worker_entity.get_id())

        run_task_server(study, args.run_mode)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module contains a class, `MerlinSpec`, which holds the unchanged
data from the Merlin specification file.

To see examples of yaml specifications, run `merlin example`.
"""
import json
import logging
import os
import shlex
from copy import deepcopy
from datetime import timedelta
from io import StringIO
from typing import Any, Dict, List, Set, TextIO, Union

import yaml
from maestrowf.datastructures.core.parameters import ParameterGenerator
from maestrowf.specification import YAMLSpecification

from merlin.spec import all_keys, defaults
from merlin.utils import find_vlaunch_var, load_array_file, needs_merlin_expansion, repr_timedelta


LOG = logging.getLogger(__name__)


# Pylint complains we have too many instance attributes but it's fine
class MerlinSpec(YAMLSpecification):  # pylint: disable=R0902
    """
    A class to represent and manage the specifications for a Merlin workflow.

    This class provides methods to verify, load, and process various sections of a
    workflow specification file, including the merlin block, batch block, and user block.
    It also handles default values and parameter mapping.

    Attributes:
        batch (Dict): A dictionary representing the batch section of the spec file.
        description (Dict): A dictionary representing the description section of the spec file.
        environment (Dict): A dictionary representing the environment section of the spec file.
        globals (Dict): A dictionary representing global parameters in the spec file.
        merlin (Dict): A dictionary representing the merlin section of the spec file.
        sections (Dict): A dictionary of all sections in the spec file.
        study (Dict): A dictionary representing the study section of the spec file.
        user (Dict): A dictionary representing the user section of the spec file.
        yaml_sections (Dict): A dictionary for YAML representation of the sections.

    Methods:
        check_section: Checks sections of the spec file for unrecognized keys.
        dump: Dumps the current spec to a pretty YAML string.
        fill_missing_defaults: Merges default values into an object.
        get_queue_list: Returns a sorted set of queues for specified steps.
        get_queue_step_relationship: Maps task queues to their associated steps.
        get_step_param_map: Creates a mapping of parameters used for each step.
        get_step_worker_map: Maps step names to associated workers.
        get_study_step_names: Returns a list of the names of the steps in the spec file.
        get_task_queues: Maps steps to their corresponding task queues.
        get_tasks_per_step: Returns the number of tasks needed for each step.
        get_worker_names: Returns a list of worker names.
        get_worker_step_map: Maps worker names to associated steps.
        load_merlin_block: Loads the merlin block from a YAML stream.
        load_spec_from_string: Creates a `MerlinSpec` object from a string (or stream) representing
            a spec file.
        load_specification: Creates a `MerlinSpec` object based on the contents of a spec file.
        load_user_block: Loads the user block from a YAML stream.
        make_queue_string: Returns a unique queue string for specified steps.
        process_spec_defaults: Fills in default values for missing sections.
        verify: Verify the spec against a valid schema.
        verify_batch_block: Validates the batch block against a predefined schema.
        verify_merlin_block: Validates the merlin block against a predefined schema.
        warn_unrecognized_keys: Checks for unrecognized keys in the spec file.
    """

    # Pylint says this call to super is useless but we'll leave it in case we want to add to __init__ in the future
    def __init__(self):  # pylint: disable=W0246
        """Initializes a MerlinSpec object."""
        super().__init__()

    @property
    def yaml_sections(self) -> Dict:
        """
        Returns a nested dictionary of all sections of the specification as used in a YAML
        specification. The structure is tailored for YAML representation.

        Returns:
            A dictionary containing the sections of the specification formatted for YAML.
        """
        return {
            "description": self.description,
            "batch": self.batch,
            "env": self.environment,
            "study": self.study,
            "global.parameters": self.globals,
            "merlin": self.merlin,
            "user": self.user,
        }

    @property
    def sections(self) -> Dict:
        """
        Returns a nested dictionary of all sections of the specification as referenced by
        [Maestro's `YAMLSpecification` class](https://maestrowf.readthedocs.io/en/latest/Maestro/reference_guide/api_reference/specification/yamlspecification.html).
        The structure is aligned with the expectations of Maestro's `YAMLSpecification` class.

        Returns:
            A dictionary containing the sections of the specification formatted for Maestro.
        """
        return {
            "description": self.description,
            "batch": self.batch,
            "environment": self.environment,
            "study": self.study,
            "globals": self.globals,
            "merlin": self.merlin,
            "user": self.user,
        }

    def __str__(self):
        env = ""
        globs = ""
        merlin = ""
        user = ""
        if self.environment:
            env = f"\n\tenvironment: \n\t\t{self.environment}"
        if self.globals:
            globs = f"\n\tglobals:\n\t\t{self.globals}"
        if self.merlin:
            merlin = f"\n\tmerlin:\n\t\t{self.merlin}"
        if self.user is not None:
            user = f"\n\tuser:\n\t\t{self.user}"
        result = f"""MERLIN SPEC OBJECT:\n\tdescription:\n\t\t{self.description}
               \n\tbatch:\n\t\t{self.batch}\n\tstudy:\n\t\t{self.study}
               {env}{globs}{merlin}{user}"""

        return result

    @classmethod
    def load_specification(cls, path: str, suppress_warning: bool = True) -> "MerlinSpec":
        """
        Load a specification file and create a `MerlinSpec` object based on its contents.

        This method reads a YAML specification file from the provided path,
        processes its contents, and returns a `MerlinSpec` object. It can also
        suppress warnings about unrecognized keys in the specification.

        Args:
            path: The path to the specification file to be loaded.
            suppress_warning: Whether to suppress warnings about unrecognized keys.

        Returns:
            A `MerlinSpec` object created from the contents of the specification file.

        Raises:
            Exception: If there is an error loading the specification file.
        """
        LOG.info("Loading specification from path: %s", path)
        try:
            # Load the YAML spec from the path
            with open(path, "r") as data:
                spec = cls.load_spec_from_string(data, needs_IO=False, needs_verification=True)
        except Exception as e:  # pylint: disable=C0103
            LOG.exception(e.args)
            raise e

        # Path not set in _populate_spec because loading spec with string
        # does not have a path so we set it here
        spec.path = path
        spec.specroot = os.path.dirname(spec.path)  # pylint: disable=W0201

        if not suppress_warning:
            spec.warn_unrecognized_keys()
        return spec

    @classmethod
    def load_spec_from_string(
        cls, string: Union[str, TextIO], needs_IO: bool = True, needs_verification: bool = False
    ) -> "MerlinSpec":  # pylint: disable=C0103
        """
        Read a specification from a string (or stream) and create a `MerlinSpec` object from it.

        This method processes a string or stream containing the specification
        and returns a `MerlinSpec` object. It can also verify the specification
        if required.

        Args:
            string: A string or stream of the specification content.
            needs_IO: Whether to treat the string as a file object.
            needs_verification: Whether to verify the specification after loading.

        Returns:
            A `MerlinSpec` object created from the provided specification content.
        """
        LOG.debug("Creating Merlin spec object...")
        # Create and populate the MerlinSpec object
        data = StringIO(string) if needs_IO else string
        spec = cls._populate_spec(data)
        spec.specroot = None  # pylint: disable=W0201
        spec.process_spec_defaults()
        LOG.debug("Merlin spec object created.")

        # Verify the spec object
        if needs_verification:
            LOG.debug("Verifying Merlin spec...")
            spec.verify()
            LOG.debug("Merlin spec verified.")

        # Convert the walltime value back to HMS if PyYAML messed with it
        for _, section in spec.yaml_sections.items():
            # Section is a list for the study block
            if isinstance(section, list):
                for step in section:
                    if "walltime" in step and isinstance(step["walltime"], int):
                        step["walltime"] = repr_timedelta(timedelta(seconds=step["walltime"]))
            # Section is a dict for all other blocks
            if isinstance(section, dict):
                if "walltime" in section and isinstance(section["walltime"], int):
                    section["walltime"] = repr_timedelta(timedelta(seconds=section["walltime"]))

        return spec

    @classmethod
    def _populate_spec(cls, data: TextIO) -> "MerlinSpec":
        """
        Helper method to load a study specification and populate its fields.

        This method reads a YAML specification from a raw text stream and
        populates the fields of a `MerlinSpec` object. It is a modified version
        of the `load_specification` method from Maestro's YAMLSpecification class,
        excluding the verification step due to compatibility issues with Maestro's schema.

        Note:
            This is basically a direct copy of YAMLSpecification's
            load_specification method from Maestro just without the call to verify.
            The verify method was breaking our code since we have no way of modifying
            Maestro's schema that they use to verify yaml files. The work around
            is to load the yaml file ourselves and create our own schema to verify
            against.

        Args:
            data: A raw text stream containing the study YAML specification data.

        Returns:
            A `MerlinSpec` object populated with information extracted from the YAML specification.
        """
        # Read in the spec file
        try:
            spec = yaml.load(data, yaml.FullLoader)
        except AttributeError:
            LOG.warning(
                "PyYAML is using an unsafe version with a known "
                "load vulnerability. Please upgrade your installation "
                "to a more recent version!"
            )
            spec = yaml.load(data, yaml.Loader)
        LOG.debug("Successfully loaded specification: \n%s", spec["description"])

        # Load in the parts of the yaml that are the same as Maestro's
        merlin_spec = cls()
        merlin_spec.path = None
        merlin_spec.description = spec.pop("description", {})
        merlin_spec.environment = spec.pop("env", {"variables": {}, "sources": [], "labels": {}, "dependencies": {}})
        merlin_spec.batch = spec.pop("batch", {})
        merlin_spec.study = spec.pop("study", [])
        merlin_spec.globals = spec.pop("global.parameters", {})

        # Reset the file pointer and load the merlin block
        data.seek(0)
        merlin_spec.merlin = MerlinSpec.load_merlin_block(data)  # pylint: disable=W0201

        # Reset the file pointer and load the user block
        data.seek(0)
        merlin_spec.user = MerlinSpec.load_user_block(data)  # pylint: disable=W0201

        return merlin_spec

    def verify(self):
        """
        Verify the specification against a valid schema.

        This method checks the current `MerlinSpec` object against a predefined
        schema to ensure that it adheres to the expected structure and
        constraints. It is similar to the verify method from Maestro's
        YAMLSpecification class but is tailored specifically for Merlin YAML
        specifications.

        Note:
            Maestro v2.0 may introduce the ability to customize the schema files
            used for verification. If that feature becomes available, then we can
            convert this file back to using Maestro's verification.

        Raises:
            Exception: If the specification does not conform to the schema,
                appropriate exceptions will be raised during the verification process.
        """
        # Load the MerlinSpec schema file
        dir_path = os.path.dirname(os.path.abspath(__file__))
        schema_path = os.path.join(dir_path, "merlinspec.json")
        with open(schema_path, "r") as json_file:
            schema = json.load(json_file)

        # Use Maestro's verification methods for shared sections
        self.verify_description(schema["DESCRIPTION"])
        self.verify_environment(schema["ENV"])
        self.verify_study(schema["STUDY_STEP"])
        self.verify_parameters(schema["PARAM"])

        # Merlin specific verification
        self.verify_merlin_block(schema["MERLIN"])
        self.verify_batch_block(schema["BATCH"])

    def get_study_step_names(self) -> List[str]:
        """
        Retrieve the names of steps in the study.

        This method iterates through the study steps and collects their names
        into a list. The returned list is unsorted.

        Returns:
            An unsorted list of strings representing the names of the
                study steps.
        """
        names = []
        for step in self.study:
            names.append(step["name"])
        return names

    def _verify_workers(self):
        """
        Verify the workers section in the Merlin block of the specification.

        This helper method checks that the steps referenced in the workers
        section of the Merlin block exist in the study steps. It raises a
        ValueError if any step specified for a worker does not match the
        defined study steps.

        Raises:
            ValueError: If a step specified in the workers section does not
                exist in the list of study step names.
        """
        # Retrieve the names of the steps in our study
        actual_steps = self.get_study_step_names()

        try:
            # Verify that the steps in merlin block's worker section actually exist
            for worker, worker_vals in self.merlin["resources"]["workers"].items():
                error_prefix = f"Problem in Merlin block with worker {worker} --"
                for step in worker_vals["steps"]:
                    if step != "all" and step not in actual_steps:
                        error_msg = (
                            f"{error_prefix} Step with the name {step}"
                            " is not defined in the study block of the yaml specification file"
                        )
                        raise ValueError(error_msg)

        except Exception:  # pylint: disable=W0706
            raise

    def verify_merlin_block(self, schema: Dict):
        """

        Verify the Merlin section of the specification file against a schema.

        This method validates the Merlin block of the specification file
        against a predefined JSON schema and verifies the workers section
        to ensure that all specified steps are defined in the study.

        Args:
            schema: The section of the predefined schema (merlinspec.json) to
                check the Merlin block against.
        """
        # Validate merlin block against the json schema
        YAMLSpecification.validate_schema("merlin", self.merlin, schema)
        # Verify the workers section within merlin block
        self._verify_workers()

    def verify_batch_block(self, schema: Dict):
        """
        Verify the batch section of the specification file against a schema.

        This method validates the batch block of the specification file
        against a predefined JSON schema and performs additional checks
        related to the walltime parameter for the LSF batch type.

        Args:
            schema: The section of the predefined schema (merlinspec.json) to
                check the batch block against.
        """
        # Validate batch block against the json schema
        YAMLSpecification.validate_schema("batch", self.batch, schema)

        # Additional Walltime checks in case the regex from the schema bypasses an error
        if self.batch["type"] == "lsf" and "walltime" in self.batch:
            LOG.warning("The walltime argument is not available in lsf.")

    @staticmethod
    def load_merlin_block(stream: TextIO) -> Dict:
        """
        Load the Merlin block from a specification file stream.

        This static method reads a YAML stream and attempts to extract
        the 'merlin' section. If the 'merlin' section is missing, it
        logs a warning and returns an empty dictionary, indicating that
        the default configuration will be used without sampling.

        Args:
            stream: A file-like object or string stream containing the
                YAML specification.

        Returns:
            The Merlin block extracted from the YAML stream. If the 'merlin'
                section is not found, an empty dictionary is returned.
        """
        try:
            merlin_block = yaml.safe_load(stream)["merlin"]
        except KeyError:
            merlin_block = {}
            warning_msg: str = (
                "Workflow specification missing \n "
                "encouraged 'merlin' section! Run 'merlin example' for examples.\n"
                "Using default configuration with no sampling."
            )
            LOG.warning(warning_msg)
        return merlin_block

    @staticmethod
    def load_user_block(stream: TextIO) -> Dict:
        """
        Load the user block from a specification file stream.

        This static method reads a YAML stream and attempts to extract
        the 'user' section. If the 'user' section is not present, it
        returns an empty dictionary.

        Args:
            stream: A file-like object or string stream containing the
                YAML specification.

        Returns:
            The user block extracted from the YAML stream. If the 'user'
                section is not found, an empty dictionary is returned.
        """
        try:
            user_block = yaml.safe_load(stream)["user"]
        except KeyError:
            user_block = {}
        return user_block

    def process_spec_defaults(self):
        """
        Fill in default values for specification sections if they are missing.

        This method iterates through the sections of the specification and
        populates any that are `None` with empty dictionaries. It then fills
        in default values for various sections, including batch, environment,
        global parameters, and step sections within the study.

        The method also handles specific cases for the VLAUNCHER variables
        in the command of each step, ensuring that default values are set
        if they are not defined by the user. Additionally, it ensures that
        workers are assigned to steps appropriately, filling in defaults
        where necessary.

        The method modifies the instance's attributes directly, ensuring that
        the specification is complete and ready for further processing.
        """
        for name, section in self.sections.items():
            if section is None:
                setattr(self, name, {})

        # fill in missing batch section defaults
        MerlinSpec.fill_missing_defaults(self.batch, defaults.BATCH["batch"])

        # fill in missing env section defaults
        MerlinSpec.fill_missing_defaults(self.environment, defaults.ENV["env"])

        # fill in missing global parameter section defaults
        MerlinSpec.fill_missing_defaults(self.globals, defaults.PARAMETER["global.parameters"])

        # fill in missing step section defaults within 'run'
        defaults.STUDY_STEP_RUN["shell"] = self.batch["shell"]
        for step in self.study:
            MerlinSpec.fill_missing_defaults(step["run"], defaults.STUDY_STEP_RUN)
            # Insert VLAUNCHER specific variables if necessary
            if "$(VLAUNCHER)" in step["run"]["cmd"]:
                SHSET = ""
                if "csh" in step["run"]["shell"]:
                    SHSET = "set "
                # We need to set default values for VLAUNCHER variables if they're not defined by the user
                for vlaunch_var, vlaunch_val in defaults.VLAUNCHER_VARS.items():
                    if not find_vlaunch_var(vlaunch_var.replace("MERLIN_", ""), step["run"]["cmd"], accept_no_matches=True):
                        # Look for predefined nodes/procs/cores/gpus values in the step and default to those
                        vlaunch_val = step["run"][vlaunch_val[0]] if vlaunch_val[0] in step["run"] else vlaunch_val[1]
                        step["run"]["cmd"] = f"{SHSET}{vlaunch_var}={vlaunch_val}\n" + step["run"]["cmd"]

        # fill in missing merlin section defaults
        MerlinSpec.fill_missing_defaults(self.merlin, defaults.MERLIN["merlin"])
        if self.merlin["resources"]["workers"] is None:
            self.merlin["resources"]["workers"] = {"default_worker": defaults.WORKER}
        else:
            # Gather a list of step names defined in the study
            all_workflow_steps = self.get_study_step_names()
            # Create a variable to track the steps assigned to workers
            worker_steps = []

            # Loop through each worker and fill in the defaults
            for _, worker_settings in self.merlin["resources"]["workers"].items():
                MerlinSpec.fill_missing_defaults(worker_settings, defaults.WORKER)
                worker_steps.extend(worker_settings["steps"])

            if "all" in worker_steps:
                steps_that_need_workers = []
            else:
                # Figure out which steps still need workers
                steps_that_need_workers = list(set(all_workflow_steps) - set(worker_steps))

            # If there are still steps remaining that haven't been assigned a worker yet,
            # assign the remaining steps to the default worker. If all the steps still need workers
            # (i.e. no workers were assigned) then default workers' steps should be "all" so we skip this
            if steps_that_need_workers and (steps_that_need_workers != all_workflow_steps):
                self.merlin["resources"]["workers"]["default_worker"] = defaults.WORKER
                self.merlin["resources"]["workers"]["default_worker"]["steps"] = steps_that_need_workers
        if self.merlin["samples"] is not None:
            MerlinSpec.fill_missing_defaults(self.merlin["samples"], defaults.SAMPLES)

        # no defaults for user block

    @staticmethod
    def fill_missing_defaults(object_to_update: Dict, default_dict: Dict):
        """
        Merge default values into an object, filling in missing attributes.

        This static method takes an object and a dictionary of default values,
        and merges the defaults into the object. It only adds missing attributes
        to the object and does not overwrite any existing attributes. If an
        attribute is present in the object but its value is `None`, it will be
        updated with the corresponding value from the defaults.

        The method works recursively, allowing for nested dictionaries.

        The method modifies the `object_to_update` in place.

        Args:
            object_to_update: The object (as a dictionary) that needs to be
                updated with default values.
            default_dict: A dictionary containing default values to merge into
                the object.

        Example:
            ```python
            >>> obj = {'a': 1, 'b': None}
            >>> defaults = {'a': 2, 'b': 3, 'c': 4}
            >>> fill_missing_defaults(obj, defaults)
            >>> print(obj)
            {'a': 1, 'b': 3, 'c': 4}
            ```
        """

        def recurse(result: Dict, recurse_defaults: Dict):
            """
            Recursively merge default values into the result object.

            This helper function checks if the current level of the `recurse_defaults`
            dictionary is a dictionary itself. If it is, it iterates through each key-value
            pair. If a key is not present in the `result` or its value is `None`, it
            assigns the value from `recurse_defaults`. If the key exists and has a value,
            it recursively calls itself to handle nested dictionaries.

            The function modifies the `result` in place.

            Args:
                result: The current state of the object being updated.
                recurse_defaults: The current level of defaults to merge.
            """
            if not isinstance(recurse_defaults, dict):
                return
            for key, val in recurse_defaults.items():
                # fmt: off
                if (key not in result) or (
                    (result[key] is None) and (recurse_defaults[key] is not None)
                ):
                    result[key] = val
                else:
                    recurse(result[key], val)
                # fmt: on

        recurse(object_to_update, default_dict)

    # ***Unsure if this method is still needed after adding json schema verification***
    def warn_unrecognized_keys(self):
        """
        Check for unrecognized keys in the specification file.

        This method verifies that all keys present in the specification file
        conform to the expected structure defined by the `MerlinSpec` class.
        It checks various sections of the specification, including "description",
        "batch", "env", "global parameters", "steps", and "merlin". For each
        section, it calls the `check_section` method to ensure that the keys
        are recognized and valid according to predefined criteria.
        """
        # check description
        MerlinSpec.check_section("description", self.description, all_keys.DESCRIPTION)

        # check batch
        MerlinSpec.check_section("batch", self.batch, all_keys.BATCH)

        # check env
        MerlinSpec.check_section("env", self.environment, all_keys.ENV)

        # check parameters
        for _, contents in self.globals.items():
            MerlinSpec.check_section("global.parameters", contents, all_keys.PARAMETER)

        # check steps
        for step in self.study:
            MerlinSpec.check_section(step["name"], step, all_keys.STUDY_STEP)
            MerlinSpec.check_section(step["name"] + ".run", step["run"], all_keys.STUDY_STEP_RUN)

        # check merlin
        MerlinSpec.check_section("merlin", self.merlin, all_keys.MERLIN)
        MerlinSpec.check_section("merlin.resources", self.merlin["resources"], all_keys.MERLIN_RESOURCES)
        for worker, contents in self.merlin["resources"]["workers"].items():
            MerlinSpec.check_section("merlin.resources.workers " + worker, contents, all_keys.WORKER)
        if self.merlin["samples"]:
            MerlinSpec.check_section("merlin.samples", self.merlin["samples"], all_keys.SAMPLES)

        # user block is not checked

    @staticmethod
    def check_section(section_name: str, section: Dict, known_keys: Set[str]):
        """
        Check a section of the specification file for unrecognized keys.

        This static method compares the keys present in a specified section
        of the specification file against a set of known keys. If any keys
        are found that are not recognized, a warning is logged indicating
        the unrecognized key and the section in which it was found.

        Args:
            section_name: The name of the section being checked.
            section: The section of the specification file to validate.
            known_keys: A set of keys that are recognized as valid for
                the specified section.
        """
        diff = set(section.keys()).difference(known_keys)

        # TODO: Maybe add a check here for required keys

        for extra in diff:
            LOG.warning(f"Unrecognized key '{extra}' found in spec section '{section_name}'.")

    def dump(self) -> str:
        """
        Dump the `MerlinSpec` instance to a formatted YAML string.

        This method converts the current state of the `MerlinSpec` instance
        into a YAML formatted string. It utilizes the `_dict_to_yaml`
        method to handle the conversion and prettification of the data.
        Additionally, it ensures that the resulting YAML string is valid
        by attempting to parse it with `yaml.safe_load`. If parsing fails,
        a ValueError is raised with details about the error.

        Returns:
            A pretty formatted YAML string representation of the
                `MerlinSpec` instance.

        Raises:
            ValueError: If there is an error while parsing the YAML string.
        """
        tab = 3 * " "
        result = self._dict_to_yaml(self.yaml_sections, "", [], tab)
        while "\n\n\n" in result:
            result = result.replace("\n\n\n", "\n\n")
        try:
            yaml.safe_load(result)
        except Exception as e:  # pylint: disable=C0103
            raise ValueError(f"Error parsing provenance spec:\n{e}") from e
        return result

    def _dict_to_yaml(self, obj: Any, string: str, key_stack: List[str], tab: int) -> str:
        """
        Convert a Python object to a formatted YAML string.

        This private method handles the conversion of various Python data
        types (strings, booleans, lists, and dictionaries) into a
        formatted YAML string. It uses an if-else structure to determine
        the type of the input object and calls the appropriate processing
        methods for each type. The method also manages indentation based
        on the current level of nesting.

        Args:
            obj: The object to convert to YAML format.
            string: The current string representation being built.
            key_stack: A stack of keys representing the current level of
                nesting in the YAML structure.
            tab: The number of spaces to use for indentation.

        Returns:
            A formatted YAML string representation of the input object.
        """
        if obj is None:
            return ""

        lvl = len(key_stack) - 1

        if isinstance(obj, str):
            return self._process_string(obj, lvl, tab)
        if isinstance(obj, bool):
            return str(obj).lower()
        if isinstance(obj, list):
            return self._process_list(obj, string, key_stack, lvl, tab)
        if isinstance(obj, dict):
            return self._process_dict(obj, string, key_stack, lvl, tab)
        return obj

    def _process_string(self, obj: str, lvl: int, tab: int) -> str:
        """
        Process a string for YAML formatting in the dump method.

        This private method takes a string and formats it for inclusion
        in a YAML output. If the string contains multiple lines, it
        transforms the string into a block scalar format using the pipe
        (`|`) character, which is suitable for YAML representation.
        The indentation is adjusted based on the current level of
        nesting.

        Args:
            obj: The string to be processed.
            lvl: The current level of indentation for the YAML output.
            tab: The number of spaces to use for indentation.

        Returns:
            The formatted string ready for YAML output.
        """
        split = obj.splitlines()
        if len(split) > 1:
            obj = "|\n" + tab * (lvl + 1) + ("\n" + tab * (lvl + 1)).join(split)
        return obj

    def _process_list(
        self,
        obj: List[Any],
        string: str,
        key_stack: List[str],
        lvl: int,
        tab: int,
    ) -> str:
        """
        Process a list for YAML formatting in the dump method.

        This private method handles the conversion of a list into a
        YAML formatted string. It determines whether to use hyphens
        for list items based on the context provided by the key stack.
        The method recursively processes each element in the list and
        manages indentation based on the current level of nesting.

        Args:
            obj: The list to be processed.
            string: The current string representation being built.
            key_stack: A stack of keys representing the current
                level of nesting in the YAML structure.
            lvl: The current level of indentation for the YAML output.
            tab: The number of spaces to use for indentation.

        Returns:
            A formatted YAML string representation of the input list.
        """
        num_entries = len(obj)
        use_hyphens = key_stack[-1] in ["paths", "sources", "git", "study"] or key_stack[0] in ["user"]
        if not use_hyphens:
            string += "["
        else:
            string += "\n"
        for i, elem in enumerate(obj):
            key_stack = deepcopy(key_stack)
            key_stack.append("elem")
            if use_hyphens:
                string += (lvl + 1) * tab + "- " + str(self._dict_to_yaml(elem, "", key_stack, tab)) + "\n"
            else:
                string += str(self._dict_to_yaml(elem, "", key_stack, tab))
                if num_entries > 1 and i != len(obj) - 1:
                    string += ", "
            key_stack.pop()
        if not use_hyphens:
            string += "]"
        return string

    def _process_dict(
        self,
        obj: Dict,
        string: str,
        key_stack: List[str],
        lvl: int,
        tab: int,
    ) -> str:  # pylint: disable=R0913
        """
        Process a dictionary for YAML formatting in the dump method.

        This private method converts a dictionary into a YAML formatted
        string. It iterates over the dictionary's key-value pairs,
        formatting each pair according to YAML syntax. The method
        handles indentation and manages the key stack to maintain the
        correct nesting level in the output.

        Args:
            obj: The dictionary to be processed.
            string: The current string representation being built.
            key_stack: A stack of keys representing the current
                level of nesting in the YAML structure.
            lvl: The current level of indentation for the YAML output.
            tab: The number of spaces to use for indentation.

        Returns:
            A formatted YAML string representation of the input dictionary.
        """
        list_offset = 2 * " "
        if len(key_stack) > 0 and key_stack[-1] != "elem":
            string += "\n"
        i = 0
        for key, val in obj.items():
            key_stack = deepcopy(key_stack)
            key_stack.append(key)
            if len(key_stack) > 1 and key_stack[-2] == "elem" and i == 0:
                # string += (tab * (lvl - 1))
                string += ""
            elif "elem" in key_stack:
                string += list_offset + (tab * lvl)
            else:
                string += tab * (lvl + 1)
            string += str(key) + ": " + str(self._dict_to_yaml(val, "", key_stack, tab)) + "\n"
            key_stack.pop()
            i += 1
        return string

    def get_step_worker_map(self) -> Dict[str, List[str]]:
        """
        Create a mapping of step names to associated workers.

        This method constructs a dictionary where each key is a step name
        and the corresponding value is a list of workers assigned to that
        step. Workers can either be associated with all steps or with
        specific steps. This method serves as the inverse of the
        [`get_worker_step_map`][spec.specification.MerlinSpec.get_worker_step_map]
        method.

        Returns:
            A dictionary mapping step names to lists of worker names
                associated with each step.
        """
        steps = self.get_study_step_names()
        step_worker_map = {step_name: [] for step_name in steps}
        for worker_name, worker_val in self.merlin["resources"]["workers"].items():
            # Case 1: worker doesn't have specific steps
            if "all" in worker_val["steps"]:
                for step_name in step_worker_map:
                    step_worker_map[step_name].append(worker_name)
            # Case 2: worker has specific steps
            else:
                for step in worker_val["steps"]:
                    step_worker_map[step].append(worker_name)
        return step_worker_map

    def get_worker_step_map(self) -> Dict[str, List[str]]:
        """
        Create a mapping of worker names to associated steps.

        This method constructs a dictionary where each key is a worker name
        and the corresponding value is a list of steps that the worker is
        assigned to monitor. Workers can either be assigned to all steps or
        to specific steps. It serves as the inverse of the
        [`get_step_worker_map`][spec.specification.MerlinSpec.get_step_worker_map]
        method.

        Returns:
            A dictionary mapping worker names to lists of step names that each
                worker monitors.
        """
        worker_step_map = {}
        steps = self.get_study_step_names()
        for worker_name, worker_val in self.merlin["resources"]["workers"].items():
            # Case 1: worker doesn't have specific steps
            if "all" in worker_val["steps"]:
                worker_step_map[worker_name] = steps
            # Case 2: worker has specific steps
            else:
                worker_step_map[worker_name] = []
                for step in worker_val["steps"]:
                    worker_step_map[worker_name].append(step)
        return worker_step_map

    def get_task_queues(self, omit_tag: bool = False) -> Dict[str, str]:
        """
        Create a mapping of steps to their corresponding task queues.

        This method constructs a dictionary where each key is a step name
        and the corresponding value is the associated task queue. The
        `omit_tag` parameter allows for the optional exclusion of the Celery
        queue tag from the queue names. It serves as the inverse of the
        [`get_queue_step_relationship`][spec.specification.MerlinSpec.get_queue_step_relationship]
        method.

        Args:
            omit_tag: If True, the Celery queue tag will be omitted
                from the task queue names. Default is False.

        Returns:
            A dictionary mapping step names to their corresponding task queues.
        """
        from merlin.config.configfile import CONFIG  # pylint: disable=C0415

        steps = self.get_study_steps()
        queues = {}
        for step in steps:
            if "task_queue" in step.run and (omit_tag or CONFIG.celery.omit_queue_tag):
                queues[step.name] = step.run["task_queue"]
            elif "task_queue" in step.run:
                queues[step.name] = CONFIG.celery.queue_tag + step.run["task_queue"]
        return queues

    def get_queue_step_relationship(self) -> Dict[str, List[str]]:
        """
        Build a mapping of task queues to their associated steps.

        This method constructs a dictionary where each key is a task queue
        name and the corresponding value is a list of steps that are
        associated with that queue. It serves as the inverse of the
        [`get_task_queues`][spec.specification.MerlinSpec.get_task_queues]
        method.

        Returns:
            A dictionary mapping task queue names to lists of step names
                associated with each queue.
        """
        from merlin.config.configfile import CONFIG  # pylint: disable=C0415

        steps = self.get_study_steps()
        relationship_tracker = {}

        for step in steps:
            if "task_queue" in step.run:
                queue_name = (
                    step.run["task_queue"]
                    if CONFIG.celery.omit_queue_tag
                    else f"{CONFIG.celery.queue_tag}{step.run['task_queue']}"
                )

                if queue_name in relationship_tracker:
                    relationship_tracker[queue_name].append(step.name)
                else:
                    relationship_tracker[queue_name] = [step.name]

        return relationship_tracker

    def get_queue_list(self, steps: Union[List[str], str], omit_tag: bool = False) -> Set[str]:
        """
        Return a sorted set of queues corresponding to specified steps.

        This method retrieves a list of task queues associated with the
        given steps. If the `steps` parameter is set to ['all'], it will
        return all available queues. The `omit_tag` parameter allows for
        the optional exclusion of the Celery queue tag from the queue names.

        Args:
            steps: A list of step names or a list containing the string 'all'
                to represent all steps, or the name of a single step.
            omit_tag: If True, the Celery queue tag will be omitted from the
                task queue names.

        Returns:
            A sorted set of unique task queues corresponding to the specified
                steps.

        Raises:
            KeyError: If any of the specified steps do not exist in the
                task queues.
        """
        queues = self.get_task_queues(omit_tag=omit_tag)
        if steps[0] == "all":
            task_queues = queues.values()
        else:
            try:
                if isinstance(steps, list):
                    task_queues = [queues[step] for step in steps]
                else:
                    task_queues = [queues[steps]]
            except KeyError:
                newline = "\n"
                LOG.error(f"Invalid steps '{steps}'! Try one of these (or 'all'):\n{newline.join(queues.keys())}")
                raise
        return sorted(set(task_queues))

    def make_queue_string(self, steps: List[str]) -> str:
        """
        Return a unique queue string for the specified steps.

        This method constructs a comma-separated string of unique task
        queues associated with the provided steps. The resulting string
        is suitable for use in command-line contexts.

        Args:
            steps: A list of step names for which to generate the
                queue string.

        Returns:
            A quoted string of unique task queues, separated by commas.
        """
        queues = ",".join(set(self.get_queue_list(steps)))
        return shlex.quote(queues)

    def get_worker_names(self) -> List[str]:
        """
        Build a list of worker names.

        This method retrieves the names of all workers defined in the
        Merlin resources and returns them as a list.

        Returns:
            A list of worker names.
        """
        result = []
        for worker in self.merlin["resources"]["workers"]:
            result.append(worker)
        return result

    def get_tasks_per_step(self) -> Dict[str, int]:
        """
        Get the number of tasks needed to complete each step.

        This method calculates the number of tasks required for each
        step in the study based on the number of samples and parameters.
        It returns a dictionary where the keys are the step names and
        the values are the corresponding number of tasks required for
        that step.

        Returns:
            A dictionary mapping step names to the number of tasks
                required for each step.
        """
        # Get the number of samples used
        samples = []
        if self.merlin["samples"] and self.merlin["samples"]["file"]:
            samples = load_array_file(self.merlin["samples"]["file"])
        num_samples = len(samples)

        # Get the column labels, the parameter labels, the number of parameters, and the steps in the study
        if num_samples > 0:
            column_labels = self.merlin["samples"]["column_labels"]
        parameter_labels = list(self.get_parameters().labels.keys())
        num_params = self.get_parameters().length
        study_steps = self.get_study_steps()

        tasks_per_step = {}
        for step in study_steps:
            cmd = step.__dict__["run"]["cmd"]
            restart_cmd = step.__dict__["run"]["restart"]

            # Default number of tasks for a step is 1
            tasks_per_step[step.name] = 1

            # If this step uses parameters, we'll at least have a num_params number of tasks to complete
            if needs_merlin_expansion(cmd, restart_cmd, parameter_labels, include_sample_keywords=False):
                tasks_per_step[step.name] = num_params

            # If merlin expansion is needed with column labels, this step uses samples
            if num_samples > 0 and needs_merlin_expansion(cmd, restart_cmd, column_labels):
                tasks_per_step[step.name] *= num_samples

        return tasks_per_step

    def _create_param_maps(self, param_gen: ParameterGenerator, expanded_labels: Dict, label_param_map: Dict):
        """
        Create mappings of tokens to expanded labels and labels to parameter values.

        This private method processes a parameter generator to create two mappings:

        1. `expanded_labels`: Maps tokens to their expanded labels based on the
            provided parameter values.
        2. `label_param_map`: Maps expanded labels to their corresponding parameter
            values.

        The expected structure for the parameter block is:

        ```
        global.parameters:
            TOKEN:
                values: [param_val_1, param_val_2]
                label: label.%%
        ```

        Args:
            param_gen: A `ParameterGenerator` object from Maestro containing the
                parameter definitions.
            expanded_labels: A dictionary to store the mapping from tokens to their
                expanded labels.
            label_param_map: A dictionary to store the mapping from labels to their
                corresponding parameter values.
        """
        for token, orig_label in param_gen.labels.items():
            for param in param_gen.parameters[token]:
                expanded_label = orig_label.replace(param_gen.label_token, str(param))
                if token in expanded_labels:
                    expanded_labels[token].append(expanded_label)
                else:
                    expanded_labels[token] = [expanded_label]
                label_param_map[expanded_label] = {token: param}

    def get_step_param_map(self) -> Dict:  # pylint: disable=R0914
        """
        Create a mapping of parameters used for each step in the study.

        This method generates a mapping of parameters for each step, where each
        step may have a command (`cmd`) and a restart command (`restart_cmd`).
        The resulting mapping has a structure similar to the following:

        ```python
        step_name_with_parameters: {
            "cmd": {
                TOKEN_1: param_1_value_1,
                TOKEN_2: param_2_value_1,
            },
            "restart_cmd": {
                TOKEN_1: param_1_value_1,
                TOKEN_3: param_3_value_1,
            }
        }
        ```

        Returns:
            A dictionary mapping step names (with parameters) to their
                respective command and restart command parameter mappings.
        """
        # Get the steps and the parameters in the study
        study_steps = self.get_study_steps()
        param_gen = self.get_parameters()

        # Create maps between tokens and expanded labels, and between labels and parameter values
        expanded_labels = {}
        label_param_map = {}
        self._create_param_maps(param_gen, expanded_labels, label_param_map)

        step_param_map = {}
        for step in study_steps:
            # Get the cmd and restart cmd for the step
            cmd = step.__dict__["run"]["cmd"]
            restart_cmd = step.__dict__["run"]["restart"]

            # Get the parameters used in this step and the labels used with those parameters
            all_params_in_step = param_gen.get_used_parameters(step)
            labels_used = [expanded_labels[param] for param in sorted(all_params_in_step)]

            # Zip all labels used for the step together (since this is how steps are named in Maestro)
            for labels in zip(*labels_used):
                # Initialize the entry in the step param map
                param_str = ".".join(labels)
                step_name_with_params = f"{step.name}_{param_str}"
                step_param_map[step_name_with_params] = {"cmd": {}, "restart_cmd": {}}

                # Populate the entry in the step param map based on which token is found in which command (cmd or restart)
                for label in labels:
                    for token, param_value in label_param_map[label].items():
                        full_token = f"{param_gen.token}({token})"
                        if full_token in cmd:
                            step_param_map[step_name_with_params]["cmd"][token] = param_value
                        if full_token in restart_cmd:
                            step_param_map[step_name_with_params]["restart_cmd"][token] = param_value

        return step_param_map



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `common` package provides shared utilities, classes, and logic used across Merlin.
It includes functionality for managing encryption, handling data sampling, working with
enumerations, and defining Celery tasks.

Subpackages:
    - `security/`: Contains functionality for managing encryption and ensuring secure
        communication. Includes modules for general encryption logic and encrypting backend traffic.

Modules:
    dumper.py: Provides functionality for dumping information to files.
    enums.py: Defines enumerations for interfaces.
    sample_index_factory.py: Houses factory methods for creating
        [`SampleIndex`][common.sample_index.SampleIndex] objects.
    sample_index.py: Implements the logic for managing the sample hierarchy, including
        the [`SampleIndex`][common.sample_index.SampleIndex] class.
    tasks.py: Defines Celery tasks, breaking down the Directed Acyclic Graph ([`DAG`][study.dag.DAG])
        into smaller tasks that Celery can manage.
    util_sampling.py: Contains utility functions for data sampling.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Merlin CLI server command module.

This module defines the `ServerCommand` class, which provides subcommands
for managing the Merlin server components such as initialization, starting,
stopping, restarting, and configuring the server. These subcommands are integrated
into the Merlin CLI via `argparse`.
"""

# pylint: disable=duplicate-code

import logging
import os
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace

from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.server.server_commands import config_server, init_server, restart_server, start_server, status_server, stop_server


LOG = logging.getLogger("merlin")


class ServerCommand(CommandEntryPoint):
    """
    Handles `server` CLI commands for interacting with Merlin's containerized server.

    Methods:
        add_parser: Adds the `server` command and its subcommands to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def _add_config_subcommand(self, server_commands: ArgumentParser):
        """
         Add the `config` subcommand to the server command parser.

        Parameters:
            server_commands (ArgumentParser): The server subparser to which the config command will be added.
        """
        server_config: ArgumentParser = server_commands.add_parser(
            "config",
            help="Making configurations for to the merlin server instance.",
            description="Config server.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        server_config.add_argument(
            "-ip",
            "--ipaddress",
            action="store",
            type=str,
            help="Set the binded IP address for the merlin server container.",
        )
        server_config.add_argument(
            "-p",
            "--port",
            action="store",
            type=int,
            help="Set the binded port for the merlin server container.",
        )
        server_config.add_argument(
            "-pwd",
            "--password",
            action="store",
            type=str,
            help="Set the password file to be used for merlin server container.",
        )
        server_config.add_argument(
            "--add-user",
            action="store",
            nargs=2,
            type=str,
            help="Create a new user for merlin server instance. (Provide both username and password)",
        )
        server_config.add_argument("--remove-user", action="store", type=str, help="Remove an exisiting user.")
        server_config.add_argument(
            "-d",
            "--directory",
            action="store",
            type=str,
            help="Set the working directory of the merlin server container.",
        )
        server_config.add_argument(
            "-ss",
            "--snapshot-seconds",
            action="store",
            type=int,
            help="Set the number of seconds merlin server waits before checking if a snapshot is needed.",
        )
        server_config.add_argument(
            "-sc",
            "--snapshot-changes",
            action="store",
            type=int,
            help="Set the number of changes that are required to be made to the merlin server before a snapshot is made.",
        )
        server_config.add_argument(
            "-sf",
            "--snapshot-file",
            action="store",
            type=str,
            help="Set the snapshot filename for database dumps.",
        )
        server_config.add_argument(
            "-am",
            "--append-mode",
            action="store",
            type=str,
            help="The appendonly mode to be set. The avaiable options are always, everysec, no.",
        )
        server_config.add_argument(
            "-af",
            "--append-file",
            action="store",
            type=str,
            help="Set append only filename for merlin server container.",
        )

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `server` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `server` command parser will be added.
        """
        server: ArgumentParser = subparsers.add_parser(
            "server",
            help="Manage broker and results server for merlin workflow.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        server.set_defaults(func=self.process_command)

        server_commands: ArgumentParser = server.add_subparsers(dest="commands")

        # `merlin server init` subcommand
        server_commands.add_parser(
            "init",
            help="Initialize merlin server resources.",
            description="Initialize merlin server",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )

        # `merlin server status` subcommand
        server_commands.add_parser(
            "status",
            help="View status of the current server containers.",
            description="View status",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )

        # `merlin server start` subcommand
        server_commands.add_parser(
            "start",
            help="Start a containerized server to be used as an broker and results server.",
            description="Start server",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )

        # `merlin server stop` subcommand
        server_commands.add_parser(
            "stop",
            help="Stop an instance of redis containers currently running.",
            description="Stop server.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )

        # `merlin server restart` subcommand
        server_commands.add_parser(
            "restart",
            help="Restart merlin server instance",
            description="Restart server.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )

        # `merlin server config` subcommand
        self._add_config_subcommand(server_commands)

    def process_command(self, args: Namespace):
        """
        Route to the appropriate server function based on the command
        specified via the CLI.

        This function processes commands related to server management,
        directing the flow to the corresponding function for actions such
        as initializing, starting, stopping, checking status, restarting,
        or configuring the server.

        Args:
            args: Parsed command-line arguments, which includes:\n
                - `commands`: The server management command to execute.
                Possible values are:
                    - `init`: Initialize the server.
                    - `start`: Start the server.
                    - `stop`: Stop the server.
                    - `status`: Check the server status.
                    - `restart`: Restart the server.
                    - `config`: Configure the server.
        """
        try:
            lc_all_val = os.environ["LC_ALL"]
            if lc_all_val != "C":
                raise ValueError(
                    f"The 'LC_ALL' environment variable is currently set to {lc_all_val} but it must be set to 'C'."
                )
        except KeyError:
            LOG.debug("The 'LC_ALL' environment variable was not set. Setting this to 'C'.")
            os.environ["LC_ALL"] = "C"  # Necessary for Redis to configure LOCALE

        if args.commands == "init":
            init_server()
        elif args.commands == "start":
            start_server()
        elif args.commands == "stop":
            stop_server()
        elif args.commands == "status":
            status_server()
        elif args.commands == "restart":
            restart_server()
        elif args.commands == "config":
            config_server(args)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for monitoring Merlin workflows and maintaining active allocations.

This module provides the `MonitorCommand` class, which implements the `monitor`
subcommand in the Merlin CLI. The purpose of the `monitor` command is to
periodically check the status of workflow tasks and worker activity to ensure
the allocation (e.g., on a computing cluster) remains alive while jobs are
in progress.
"""

# pylint: disable=duplicate-code

import logging
import time
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter

from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.cli.utils import get_merlin_spec_with_override
from merlin.monitor.monitor import Monitor
from merlin.router import check_merlin_status


LOG = logging.getLogger("merlin")


class MonitorCommand(CommandEntryPoint):
    """
    Handles `monitor` CLI command for monitoring workflows to ensure the allocation remains alive.

    Methods:
        add_parser: Adds the `monitor` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `monitor` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `monitor` command parser will be added.
        """
        monitor: ArgumentParser = subparsers.add_parser(
            "monitor",
            help="Check for active workers on an allocation.",
            formatter_class=RawTextHelpFormatter,
        )
        monitor.set_defaults(func=self.process_command)
        monitor.add_argument("specification", type=str, help="Path to a Merlin YAML spec file")
        monitor.add_argument(
            "--steps",
            nargs="+",
            type=str,
            dest="steps",
            default=["all"],
            help="The specific steps (tasks on the server) in the YAML file defining the queues you want to monitor",
        )
        monitor.add_argument(
            "--vars",
            action="store",
            dest="variables",
            type=str,
            nargs="+",
            default=None,
            help="Specify desired Merlin variable values to override those found in the specification. Space-delimited. "
            "Example: '--vars LEARN=path/to/new_learn.py EPOCHS=3'",
        )
        monitor.add_argument(
            "--task_server",
            type=str,
            default="celery",
            help="Task server type for which to monitor the workers.\
                                Default: %(default)s",
        )
        monitor.add_argument(
            "--sleep",
            type=int,
            default=60,
            help="Sleep duration between checking for workers.\
                                        Default: %(default)s",
        )
        monitor.add_argument(
            "--no-restart",
            "-n",
            action="store_true",
            help="Disable the automatic restart functionality for this monitor.",
        )

    def process_command(self, args: Namespace):
        """
        CLI command to monitor Merlin workers and queues to maintain
        allocation status.

        This function periodically checks the status of Merlin workers and
        the associated queues to ensure that the allocation remains active.
        It includes a sleep interval to wait before each check, including
        the initial one.

        Args:
            args: Parsed command-line arguments, which may include:\n
                - `sleep`: The duration (in seconds) to wait before
                checking the queue status again.
        """
        spec, _ = get_merlin_spec_with_override(args)

        # Give the user time to queue up jobs in case they haven't already
        time.sleep(args.sleep)

        if args.steps != ["all"]:
            LOG.warning(
                "The `--steps` argument of the `merlin monitor` command is set to be deprecated in Merlin v1.14 "
                "For now, using this argument will tell merlin to use the version of the monitor command from Merlin v1.12."
            )
            # Check if we still need our allocation
            while check_merlin_status(args, spec):
                LOG.info("Monitor: found tasks in queues and/or tasks being processed")
                time.sleep(args.sleep)
        else:
            monitor = Monitor(spec, args.sleep, args.task_server, args.no_restart)
            monitor.monitor_all_runs()

        LOG.info("Monitor: ... stop condition met")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for displaying configuration and environment information.

This module defines the `InfoCommand` class, which handles the `info` subcommand
of the Merlin CLI. The `info` command is intended to display detailed information
about the current Merlin configuration, Python environment, and other diagnostic
data useful for debugging or verifying setup.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentParser, Namespace

from merlin.cli.commands.command_entry_point import CommandEntryPoint


LOG = logging.getLogger("merlin")


class InfoCommand(CommandEntryPoint):
    """
    Handles `info` CLI command for viewing information about server connections.

    Methods:
        add_parser: Adds the `info` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `info` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `info` command parser will be added.
        """
        info: ArgumentParser = subparsers.add_parser(
            "info",
            help="display info about the merlin configuration and the python configuration. Useful for debugging.",
        )
        info.set_defaults(func=self.process_command)

    def process_command(self, args: Namespace):
        """
        CLI command to print merlin configuration info.

        Args:
            args: Parsed CLI arguments.
        """
        # if this is moved to the toplevel per standard style, merlin is unable to generate the (needed) default config file
        from merlin import display  # pylint: disable=import-outside-toplevel

        display.print_info(args)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""Updated celery configuration."""
from __future__ import absolute_import, print_function

import logging
import os
from typing import Any, Dict, List, Optional, Union

import billiard
import celery
import psutil
from celery import Celery, states
from celery.backends.redis import RedisBackend  # noqa: F401 ; Needed for celery patch
from celery.signals import celeryd_init, worker_process_init, worker_shutdown

import merlin.common.security.encrypt_backend_traffic
from merlin.common.enums import WorkerStatus
from merlin.config import broker, celeryconfig, results_backend
from merlin.config.configfile import CONFIG
from merlin.config.utils import Priority, get_priority
from merlin.db_scripts.merlin_db import MerlinDatabase
from merlin.utils import nested_namespace_to_dicts


LOG: logging.Logger = logging.getLogger(__name__)


def patch_celery():
    """
    Patch redis backend so that errors in chords don't break workflows.
    Celery has error callbacks but they do not work properly on chords that
    are nested within chains.

    Credit to this function goes to
    [the following post](https://danidee10.github.io/2019/07/09/celery-chords.html).
    """

    def _unpack_chord_result(
        self,
        tup,
        decode,
        EXCEPTION_STATES=states.EXCEPTION_STATES,
        PROPAGATE_STATES=states.PROPAGATE_STATES,
    ):
        _, tid, state, retval = decode(tup)

        if state in EXCEPTION_STATES:
            retval = self.exception_to_python(retval)
        if state in PROPAGATE_STATES:
            # retval is an Exception
            retval = f"{retval.__class__.__name__}: {str(retval)}"

        return retval

    celery.backends.redis.RedisBackend._unpack_chord_result = _unpack_chord_result

    return celery


# This function has to have specific args/return values for celery so ignore pylint
def route_for_task(
    name: str,
    args: List[Any],
    kwargs: Dict[Any, Any],
    options: Dict[Any, Any],
    task: celery.Task = None,
    **kw: Dict[Any, Any],
) -> Dict[Any, Any]:  # pylint: disable=W0613,R1710
    """
    Custom task router for Celery queues.

    This function routes tasks to specific queues based on the task name.
    If the task name contains a colon, it splits the name to determine the queue.

    Args:
        name: The name of the task being routed.
        args: The positional arguments passed to the task.
        kwargs: The keyword arguments passed to the task.
        options: Additional options for the task.
        task: The task instance (default is None).
        **kw: Additional keyword arguments for THIS function (not the task).

    Returns:
        A dictionary specifying the queue to route the task to.
            If the task name contains a colon, it returns a dictionary with
            the key "queue" set to the queue name. Otherwise, it returns
            an empty dictionary.

    Example:
        Using a colon in the name will return the string before the colon as the queue:

        ```python
        >>> route_for_task("my_queue:my_task")
        {"queue": "my_queue"}
        ```
    """
    if ":" in name:
        queue, _ = name.split(":")
        return {"queue": queue}


merlin.common.security.encrypt_backend_traffic.set_backend_funcs()


BROKER_SSL: bool = True
RESULTS_SSL: bool = False
BROKER_URI: Optional[str] = ""
RESULTS_BACKEND_URI: Optional[str] = ""
try:
    BROKER_URI = broker.get_connection_string()
    sanitized_broker_uri = broker.get_connection_string(include_password=False)
    LOG.debug(f"Broker connection string: {sanitized_broker_uri}.")
    BROKER_SSL = broker.get_ssl_config()
    LOG.debug(f"Broker SSL {'enabled' if BROKER_SSL else 'disabled'}.")
    RESULTS_BACKEND_URI = results_backend.get_connection_string()
    sanitized_results_backend_uri = results_backend.get_connection_string(include_password=False)
    LOG.debug(f"Results backend connection string: {sanitized_results_backend_uri}.")
    RESULTS_SSL = results_backend.get_ssl_config(celery_check=True)
    LOG.debug(f"Results backend SSL {'enabled' if RESULTS_SSL else 'disabled'}.")
except ValueError:
    # These variables won't be set if running with '--local'.
    BROKER_URI = None
    RESULTS_BACKEND_URI = None

app_name = "merlin_test_app" if os.getenv("CELERY_ENV") == "test" else "merlin"

# initialize app with essential properties
app: Celery = patch_celery().Celery(
    app_name,
    broker=BROKER_URI,
    backend=RESULTS_BACKEND_URI,
    broker_use_ssl=BROKER_SSL,
    redis_backend_use_ssl=RESULTS_SSL,
    task_routes=(route_for_task,),
)

# set task priority defaults to prioritize workflow tasks over task-expansion tasks
task_priority_defaults: Dict[str, Union[int, Priority]] = {
    "task_queue_max_priority": 10,
    "task_default_priority": get_priority(Priority.MID),
}
if CONFIG.broker.name.lower() == "redis":
    app.conf.broker_transport_options = {
        "priority_steps": list(range(1, 11)),
        "queue_order_strategy": "priority",
    }
app.conf.update(**task_priority_defaults)

# load merlin config defaults
app.conf.update(**celeryconfig.DICT)

# load config overrides from app.yaml
if (
    not hasattr(CONFIG.celery, "override")
    or (CONFIG.celery.override is None)
    or (not nested_namespace_to_dicts(CONFIG.celery.override))  # only true if len == 0
):
    LOG.debug("Skipping celery config override; 'celery.override' field is empty.")
else:
    override_dict: Dict = nested_namespace_to_dicts(CONFIG.celery.override)
    override_str: str = ""
    i: int = 0
    for k, v in override_dict.items():
        if k not in str(app.conf.__dict__):
            raise ValueError(f"'{k}' is not a celery configuration.")
        override_str += f"\t{k}:\t{v}"
        if i != len(override_dict) - 1:
            override_str += "\n"
        i += 1
    LOG.info(
        "Overriding default celery config with 'celery.override' in 'app.yaml':\n%s",
        override_str,
    )
    app.conf.update(**override_dict)

# auto-discover tasks
app.autodiscover_tasks(["merlin.common"])


@celeryd_init.connect
def handle_worker_startup(sender: str = None, **kwargs):
    """
    Store information about each physical worker instance in the database.

    When workers first start up, the `celeryd_init` signal is the first signal
    that they receive. This specific function will create a
    [`PhysicalWorkerModel`][db_scripts.data_models.PhysicalWorkerModel] and
    store it in the database. It does this through the use of the
    [`MerlinDatabase`][db_scripts.merlin_db.MerlinDatabase] class.

    Args:
        sender (str): The hostname of the worker that was just started
    """
    if sender is not None:
        LOG.debug(f"Worker {sender} has started.")
        options = kwargs.get("options", None)
        if options is not None:
            try:
                # Sender name is of the form celery@worker_name.%hostname
                worker_name, host = sender.split("@")[1].split(".%")
                merlin_db = MerlinDatabase()
                logical_worker = merlin_db.get("logical_worker", worker_name=worker_name, queues=options.get("queues"))
                physical_worker = merlin_db.create(
                    "physical_worker",
                    name=str(sender),
                    host=host,
                    status=WorkerStatus.RUNNING,
                    logical_worker_id=logical_worker.get_id(),
                    pid=os.getpid(),
                )
                logical_worker.add_physical_worker(physical_worker.get_id())
            # Without this exception catcher, celery does not output any errors that happen here
            except Exception as exc:
                LOG.error(f"An error occurred when processing handle_worker_startup: {exc}")
        else:
            LOG.warning("On worker connect could not retrieve worker options from Celery.")
    else:
        LOG.warning("On worker connect no sender was provided from Celery.")


@worker_shutdown.connect
def handle_worker_shutdown(sender: str = None, **kwargs):
    """
    Update the database for a worker entry when a worker shuts down.

    Args:
        sender (str): The hostname of the worker that was just started
    """
    if sender is not None:
        LOG.debug(f"Worker {sender} is shutting down.")
        merlin_db = MerlinDatabase()
        physical_worker = merlin_db.get("physical_worker", str(sender))
        if physical_worker:
            physical_worker.set_status(WorkerStatus.STOPPED)
            physical_worker.set_pid(None)  # Clear the pid
        else:
            LOG.warning(f"Worker {sender} not found in the database.")
    else:
        LOG.warning("On worker shutdown no sender was provided from Celery.")


# Pylint believes the args are unused, I believe they're used after decoration
@worker_process_init.connect()
def setup(**kwargs: Dict[Any, Any]):  # pylint: disable=W0613
    """
    Set affinity for the worker on startup (works on toss3 nodes).

    Args:
        **kwargs: Keyword arguments.
    """
    if "CELERY_AFFINITY" in os.environ and int(os.environ["CELERY_AFFINITY"]) > 1:
        # Number of cpus between workers.
        cpu_skip: int = int(os.environ["CELERY_AFFINITY"])
        npu: int = psutil.cpu_count()
        process: psutil.Process = psutil.Process()
        # pylint is upset that typing accesses a protected class, ignoring W0212
        # pylint is upset that billiard doesn't have a current_process() method - it does
        current: billiard.process._MainProcess = billiard.current_process()  # pylint: disable=W0212, E1101
        prefork_id: int = current._identity[0] - 1  # pylint: disable=W0212  # range 0:nworkers-1
        cpu_slot: int = (prefork_id * cpu_skip) % npu
        process.cpu_affinity(list(range(cpu_slot, cpu_slot + cpu_skip)))



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
SQLite-based generic store implementation for Merlin entities.

This module defines `SQLiteStoreBase`, a generic base class for managing
entity persistence using SQLite as the underlying storage. It provides
core CRUD operations (create, retrieve, update, delete) and dynamic table
creation based on model class field definitions.

This module is intended to be subclassed by entity-specific store classes
in the Merlin backend architecture.

See also:
    - merlin.backends.store_base: Base class
    - merlin.backends.sqlite.sqlite_stores: Concrete store implementations
    - merlin.db_scripts.data_models: Data model definitions
"""

import logging
from datetime import datetime
from typing import Any, Generic, List, Optional, Type

from merlin.backends.sqlite.sqlite_connection import SQLiteConnection
from merlin.backends.store_base import StoreBase, T
from merlin.backends.utils import deserialize_entity, get_not_found_error_class, serialize_entity


LOG = logging.getLogger(__name__)


class SQLiteStoreBase(StoreBase[T], Generic[T]):
    """
    Base class for SQLite-based stores.

    This class provides common functionality for saving, retrieving, and deleting
    entities in a SQLite database.

    Attributes:
        table_name (str): The table name used for SQLite entries.
        model_class (Type[T]): The model class used for deserialization.

    Methods:
        save: Save or update an entity in the database.
        retrieve: Retrieve an entity from the database by ID.
        retrieve_all: Query the database for all entities of this type.
        delete: Delete an entity from the database by ID.
    """

    def __init__(self, table_name: str, model_class: Type[T]):
        """
        Initialize the SQLite store with a SQLite connection.

        Args:
            table_name: The table name used for SQLite entries.
            model_class: The model class used for deserialization.
        """
        self.table_name: str = table_name
        self.model_class: Type[T] = model_class
        self.create_table_if_not_exists()

    def _get_sqlite_type(self, py_type: Any) -> str:
        """
        Map Python types to SQLite types.

        Args:
            py_type: A Python type hint (e.g., str, int, List[str], etc.)

        Returns:
            A string representing the corresponding SQLite column type.
        """
        origin_type = getattr(py_type, "__origin__", py_type)
        result = "TEXT"  # Default fallback

        # Handle generics like List[str], Dict[str, Any], etc.
        if origin_type in (list, dict, set):
            result = "TEXT"  # store as JSON string
        elif py_type == str:
            result = "TEXT"
        elif py_type == int:
            result = "INTEGER"
        elif py_type == float:
            result = "REAL"
        elif py_type == bool:
            result = "INTEGER"  # SQLite uses 0 and 1 for booleans
        elif py_type == datetime:
            result = "TEXT"  # ISO format string

        return result

    def create_table_if_not_exists(self):
        """
        Create the table if it doesn't exist.
        """
        field_defs = []

        for field_obj in self.model_class.get_class_fields():
            col_name = field_obj.name
            col_type = self._get_sqlite_type(field_obj.type)
            field_defs.append(f"{col_name} {col_type}")

        field_defs_str = ", ".join(field_defs)

        with SQLiteConnection() as conn:
            conn.execute(f"CREATE TABLE IF NOT EXISTS {self.table_name} ({field_defs_str});")

    def save(self, entity: T):
        """
        Save or update an entity in the SQLite database.

        Args:
            entity: The entity to save.
        """
        # Try to retrieve any existing entity with this ID
        existing_data = self.retrieve(entity.id)

        # If the entity already exists, update it
        if existing_data:
            LOG.debug(f"Attempting to update {self.table_name} with id '{entity.id}'...")
            existing_data.update_fields(entity.to_dict())
            serialized_data = serialize_entity(existing_data)
            set_str = ", ".join(
                f"{field.name} = :{field.name}" for field in self.model_class.get_class_fields() if field.name != "id"
            )
            with SQLiteConnection() as conn:
                conn.execute(
                    f"""
                    UPDATE {self.table_name}
                    SET {set_str}
                    WHERE id = :id
                """,
                    serialized_data,
                )
            LOG.debug(f"Successfully updated {self.table_name} with id '{entity.id}'.")
        # If the entity does not already exist, create it
        else:
            LOG.debug(f"Creating a {self.table_name} entry in SQLite...")
            serialized_data = serialize_entity(entity)
            fields = [field.name for field in self.model_class.get_class_fields()]
            columns_str = ", ".join(fields)
            placeholders_str = ", ".join(f":{name}" for name in fields)
            with SQLiteConnection() as conn:
                conn.execute(
                    f"""
                    INSERT INTO {self.table_name} ({columns_str})
                    VALUES ({placeholders_str})
                """,
                    serialized_data,
                )
            LOG.debug(f"Successfully created a {self.table_name} with id '{entity.id}' in SQLite.")

    def retrieve(self, identifier: str, by_name: bool = False) -> Optional[T]:
        """
        Retrieve an entity from the SQLite database by ID or name.

        Args:
            identifier: The ID or name of the entity to retrieve.
            by_name: If True, interpret the identifier as a name. If False, interpret it as an ID.

        Returns:
            The entity if found, None otherwise.
        """
        LOG.debug(f"Retrieving identifier {identifier} in SQLiteStoreBase.")

        id_type = "name" if by_name else "id"

        with SQLiteConnection() as conn:
            cursor = conn.execute(f"SELECT * FROM {self.table_name} WHERE {id_type} = :identifier", {"identifier": identifier})
            row = cursor.fetchone()

            if row is None:
                return None

            return deserialize_entity(dict(row), self.model_class)

    def retrieve_all(self) -> List[T]:
        """
        Query the SQLite database for all entities of this type.

        Returns:
            A list of entities.
        """
        entity_type = f"{self.table_name}s" if self.table_name != "study" else "studies"
        LOG.info(f"Fetching all {entity_type} from SQLite...")

        with SQLiteConnection() as conn:
            cursor = conn.execute(f"SELECT * FROM {self.table_name}")
            all_entities = []

            for row in cursor.fetchall():
                try:
                    entity = deserialize_entity(dict(row), self.model_class)
                    if entity:
                        all_entities.append(entity)
                    else:
                        LOG.warning(
                            f"{self.table_name.capitalize()} with id '{row['id']}' could not be retrieved or does not exist."
                        )
                except Exception as exc:  # pylint: disable=broad-except
                    LOG.error(f"Error retrieving {self.table_name} with id '{row['id']}': {exc}")

        LOG.info(f"Successfully retrieved {len(all_entities)} {entity_type} from SQLite.")
        return all_entities

    def delete(self, identifier: str, by_name: bool = False):
        """
        Delete an entity from the SQLite database by ID or name.

        Args:
            identifier: The ID or name of the entity to delete.
            by_name: If True, interpret the identifier as a name. If False, interpret it as an ID.
        """
        id_type = "name" if by_name else "id"
        LOG.info(f"Attempting to delete {self.table_name} with {id_type} '{identifier}' from SQLite...")

        entity = self.retrieve(identifier, by_name=by_name)
        if entity is None:
            error_class = get_not_found_error_class(self.model_class)
            raise error_class(f"{self.table_name.capitalize()} with {id_type} '{identifier}' does not exist in the database.")

        # Delete the entity
        with SQLiteConnection() as conn:
            cursor = conn.execute(f"DELETE FROM {self.table_name} WHERE {id_type} = :identifier", {"identifier": identifier})

            if cursor.rowcount == 0:
                LOG.warning(f"No rows were deleted for {self.table_name} with {id_type} '{identifier}'")
            else:
                LOG.info(f"Successfully deleted {self.table_name} '{identifier}' from SQLite.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing entities and their associated runs in the database.

This module provides a mixin class, `RunManagementMixin`, designed to simplify the management
of runs associated with an entity. The mixin can be used by any class that has the necessary
attributes and methods to support run management, such as `reload_data`, `save`, and an
`entity_info` object containing a `runs` list.
"""

from typing import List


class RunManagementMixin:
    """
    Mixin for managing runs associated with an entity.

    This mixin provides utility methods for handling run IDs associated with an entity.
    It assumes that the class using this mixin has the necessary attributes and methods
    to support run management, including `reload_data`, `save`, and an `entity_info` object
    containing a `runs` list.

    Methods:
        get_runs:
            Retrieve the IDs of the runs associated with the entity.

        add_run:
            Add a run ID to the list of runs.

        remove_run:
            Remove a run ID from the list of runs.
    """

    def get_runs(self) -> List[str]:
        """
        Get every run listed in this entity.

        Assumptions:
            - The class using this must have a `reload_data` method
            - The class using this must have an `entity_info` object containing a `runs` list

        Returns:
            A list of run IDs.
        """
        self.reload_data()
        return self.entity_info.runs

    def add_run(self, run_id: str):
        """
        Add a new run ID to the list of runs.

        Assumptions:
            - The class using this must have an `entity_info` object containing a `runs` list
            - The class using this must have a `save` method

        Args:
            run_id: The ID of the run to add.
        """
        self.entity_info.runs.append(run_id)
        self.save()

    def remove_run(self, run_id: str):
        """
        Remove a run ID from the list of runs.

        Does *not* delete the run entity from the database. This will only remove the run's ID
        from the list in this entity.

        Assumptions:
            - The class using this must have a `reload_data` method
            - The class using this must have an `entity_info` object containing a `runs` list
            - The class using this must have a `save` method

        Args:
            run_id: The ID of the run to remove.
        """
        self.reload_data()
        self.entity_info.runs.remove(run_id)
        self.save()

    def construct_run_string(self) -> str:
        """
        Constructs and returns a formatted string representation of all runs associated
        with the current instance.

        Returns:
            A formatted string containing details of all runs.
        """
        from merlin.db_scripts.entities.run_entity import RunEntity  # pylint: disable=import-outside-toplevel

        runs = self.get_runs()
        run_str = ""
        if runs:
            for run_id in runs:
                try:
                    run = RunEntity.load(run_id, self.backend)
                    run_str += f"  - ID: {run.get_id()}\n    Workspace: {run.get_workspace()}\n"
                except Exception:  # pylint: disable=broad-exception-caught
                    run_str += f"  - ID: {run_id} (Error loading run)\n"
        else:
            run_str = "  No runs found.\n"
        return run_str



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing task queues associated with entities.

This module provides a mixin class, `QueueManagementMixin`, designed to simplify the
management of task queues associated with an entity. The mixin can be used by any
class that has the necessary attributes to support queue retrieval, specifically an
`entity_info` object containing a `queues` list.
"""

from typing import List


# pylint: disable=too-few-public-methods


class QueueManagementMixin:
    """
    Mixin for managing queues associated with an entity.

    This mixin provides a method for retrieving the task queues assigned to an entity.
    It assumes that the class using this mixin has an `entity_info` object containing
    a `queues` list.

    Methods:
        get_queues:
            Retrieve the task queues assigned to the entity.
    """

    def get_queues(self) -> List[str]:
        """
        Get the queues that this entity is assigned to.

        Assumptions:
            - The class using this must have an `entity_info` object containing a `queues` list

        Returns:
            A list of queue names.
        """
        return self.entity_info.queues



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing database entities related to logical workers.

This module defines the `LogicalWorkerEntity` class, which extends the abstract base class
[`DatabaseEntity`][db_scripts.entities.db_entity.DatabaseEntity], to encapsulate logical-worker-specific
operations and behaviors.
"""

import logging
from typing import List

from merlin.db_scripts.data_models import LogicalWorkerModel
from merlin.db_scripts.entities.db_entity import DatabaseEntity
from merlin.db_scripts.entities.mixins.name import NameMixin
from merlin.db_scripts.entities.mixins.queue_management import QueueManagementMixin
from merlin.db_scripts.entities.mixins.run_management import RunManagementMixin
from merlin.db_scripts.entities.physical_worker_entity import PhysicalWorkerEntity


LOG = logging.getLogger("merlin")


class LogicalWorkerEntity(DatabaseEntity[LogicalWorkerModel], RunManagementMixin, QueueManagementMixin, NameMixin):
    """
    A class representing a logical worker in the database.

    This class provides methods to interact with and manage a logical worker's data, including
    retrieving, adding, and removing run IDs and physical worker IDs from their respective lists,
    as well as saving or deleting the logical worker itself from the database.

    Attributes:
        entity_info (db_scripts.data_models.LogicalWorkerModel): An instance of the `LogicalWorkerModel`
            class containing the logical worker's metadata.
        backend (backends.results_backend.ResultsBackend): An instance of the `ResultsBackend`
            class used to interact with the database.

    Methods:
        __repr__:
            Provide a string representation of the `LogicalWorkerEntity` instance.

        __str__:
            Provide a human-readable string representation of the `LogicalWorkerEntity` instance.

        reload_data:
            Reload the latest data for this logical worker from the database.

        get_id:
            Retrieve the unique ID of the logical worker. _Implementation found in
                [`DatabaseEntity.get_id`][db_scripts.entities.db_entity.DatabaseEntity.get_id]._

        get_additional_data:
            Retrieve any additional metadata associated with the logical worker. _Implementation found in
                [`DatabaseEntity.get_additional_data`][db_scripts.entities.db_entity.DatabaseEntity.get_additional_data]._

        get_name:
            Retrieve the name of the logical worker.

        get_runs:
            Retrieve the IDs of the runs using this logical worker.

        add_run:
            Add a run ID to the list of runs.

        remove_run:
            Remove a run ID from the list of runs.

        get_physical_workers:
            Retrieve the IDs of the physical workers created from this logical worker.

        add_physical_worker:
            Add a physical worker ID to the list of physical workers.

        remove_physical_worker:
            Remove a physical worker ID from the list of physical workers.

        get_queues:
            Retrieve the list of queues that this worker is assigned to.

        save:
            Save the current state of the logical worker to the database.

        load:
            (classmethod) Load a `LogicalWorkerEntity` instance from the database by its ID.

        delete:
            (classmethod) Delete a logical worker from the database by its ID.
    """

    @classmethod
    def _get_entity_type(cls) -> str:
        return "logical_worker"

    def __repr__(self) -> str:
        """
        Provide a string representation of the `LogicalWorkerEntity` instance.

        Returns:
            A human-readable string representation of the `LogicalWorkerEntity` instance.
        """
        return (
            f"LogicalWorkerEntity("
            f"id={self.get_id()}, "
            f"name={self.get_name()}, "
            f"runs={self.get_runs()}, "
            f"queues={self.get_queues()}, "
            f"physical_workers={self.get_physical_workers()}, "
            f"additional_data={self.get_additional_data()}, "
            f"backend={self.backend.get_name()})"
        )

    def __str__(self) -> str:
        """
        Provide a string representation of the `LogicalWorkerEntity` instance.

        Returns:
            A human-readable string representation of the `LogicalWorkerEntity` instance.
        """
        worker_id = self.get_id()
        physical_workers = [
            PhysicalWorkerEntity.load(physical_worker_id, self.backend) for physical_worker_id in self.get_physical_workers()
        ]
        physical_worker_str = ""
        if physical_workers:
            for physical_worker in physical_workers:
                physical_worker_str += f"  - ID: {physical_worker.get_id()}\n    Name: {physical_worker.get_name()}\n"
        else:
            physical_worker_str = "  No physical workers found.\n"
        return (
            f"Logical Worker with ID {worker_id}\n"
            f"------------{'-' * len(worker_id)}\n"
            f"Name: {self.get_name()}\n"
            f"Runs:\n{self.construct_run_string()}"
            f"Queues: {self.get_queues()}\n"
            f"Physical Workers:\n{physical_worker_str}"
            f"Additional Data: {self.get_additional_data()}\n\n"
        )

    def get_physical_workers(self) -> List[str]:
        """
        Get the physical instances of this logical worker.

        Returns:
            A list of physical worker IDs.
        """
        self.reload_data()
        return self.entity_info.physical_workers

    def add_physical_worker(self, physical_worker_id: str):
        """
        Add a new physical worker id to the list of physical workers.

        Args:
            physical_worker_id: The id of the physical worker to add.
        """
        self.entity_info.physical_workers.append(physical_worker_id)
        self.save()

    def remove_physical_worker(self, physical_worker_id: str):
        """
        Remove a physical worker id from the list of physical workers.

        Does *not* delete a [`PhysicalWorkerEntity`][db_scripts.entities.physical_worker_entity.PhysicalWorkerEntity]
        from the database. This will only remove the physical worker's id from the list in this worker.

        Args:
            physical_worker_id: The ID of the physical worker to remove.
        """
        self.reload_data()
        self.entity_info.physical_workers.remove(physical_worker_id)
        self.save()



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module contains the functionality necessary to interact with everything
stored in Merlin's database.
"""

import logging
from typing import Any, Dict, List

from merlin.backends.backend_factory import backend_factory
from merlin.backends.results_backend import ResultsBackend
from merlin.db_scripts.entity_managers.entity_manager import EntityManager
from merlin.db_scripts.entity_managers.logical_worker_manager import LogicalWorkerManager
from merlin.db_scripts.entity_managers.physical_worker_manager import PhysicalWorkerManager
from merlin.db_scripts.entity_managers.run_manager import RunManager
from merlin.db_scripts.entity_managers.study_manager import StudyManager
from merlin.exceptions import EntityManagerNotSupportedError


LOG = logging.getLogger("merlin")


class MerlinDatabase:
    """
    High-level interface for accessing Merlin database entities.

    This class provides a unified interface to all entity managers in Merlin.

    Attributes:
        backend (backends.results_backend.ResultsBackend): A `ResultsBackend` instance.
        logical_workers (db_scripts.entity_managers.logical_worker_manager.LogicalWorkerManager):
            A `LogicalWorkerManager` instance.
        physical_workers (db_scripts.entity_managers.physical_worker_manager.PhysicalWorkerManager):
            A `PhysicalWorkerManager` instance.
        runs (db_scripts.entity_managers.run_manager.RunManager): A `RunManager` instance.
        studies (db_scripts.entity_managers.study_manager.StudyManager): A `StudyManager` instance.

    Methods:
        get_db_type: Retrieve the type of the backend being used (e.g., Redis, SQL).
        get_db_version: Retrieve the version of the backend.
        get_connection_string: Retrieve the backend connection string.
        create: Create a new entity of the specified type.
        get: Get an entity by type and identifier.
        get_all:  Get all entities of a specific type.
        delete: Delete an entity by type and identifier.
        delete_all: Delete all entities of a specific type.
        get_everything: Get all entities from all entity managers.
        delete_everything: Delete all entities from all entity managers.
    """

    def __init__(self):
        """
        Initialize a new MerlinDatabase instance.
        """
        from merlin.config.configfile import CONFIG  # pylint: disable=import-outside-toplevel

        self.backend: ResultsBackend = backend_factory.get_backend(CONFIG.results_backend.name.lower())
        self._entity_managers: Dict[str, EntityManager] = {
            "study": StudyManager(self.backend),
            "run": RunManager(self.backend),
            "logical_worker": LogicalWorkerManager(self.backend),
            "physical_worker": PhysicalWorkerManager(self.backend),
        }

        # Set up cross-references for managers that need them
        for manager in self._entity_managers.values():
            if hasattr(manager, "set_db_reference"):
                manager.set_db_reference(self)

    # Provide direct access to entity managers for convenience
    @property
    def studies(self) -> StudyManager:
        """
        Get the study manager.

        Returns:
            A [`StudyManager`][db_scripts.entity_managers.study_manager.StudyManager]
                instance.
        """
        return self._entity_managers["study"]

    @property
    def runs(self) -> RunManager:
        """
        Get the run manager.

        Returns:
            A [`RunManager`][db_scripts.entity_managers.run_manager.RunManager]
                instance.
        """
        return self._entity_managers["run"]

    @property
    def logical_workers(self) -> LogicalWorkerManager:
        """
        Get the logical worker manager.

        Returns:
            A [`LogicalWorkerManager`][db_scripts.entity_managers.logical_worker_manager.LogicalWorkerManager]
                instance.
        """
        return self._entity_managers["logical_worker"]

    @property
    def physical_workers(self) -> PhysicalWorkerManager:
        """
        Get the physical worker manager.

        Returns:
            A [`PhysicalWorkerManager`][db_scripts.entity_managers.physical_worker_manager.PhysicalWorkerManager]
                instance.
        """
        return self._entity_managers["physical_worker"]

    def get_db_type(self) -> str:
        """
        Retrieve the type of backend.

        Returns:
            The type of backend (e.g. redis, sql, etc.).
        """
        return self.backend.get_name()

    def get_db_version(self) -> str:
        """
        Get the version of the backend.

        Returns:
            The version number of the backend.
        """
        return self.backend.get_version()

    def get_connection_string(self) -> str:
        """
        Get the connection string to the backend.

        Returns:
            The connection string to the backend.
        """
        return self.backend.get_connection_string()

    def _validate_entity_type(self, entity_type: str):
        """
        Check to make sure the entity type passed in is supported.

        Args:
            entity_type: The type of entity to validate (study, run, logical_worker, physical_worker).
        """
        if entity_type not in self._entity_managers:
            raise EntityManagerNotSupportedError(f"Entity type not supported: {entity_type}")

    def create(self, entity_type: str, *args, **kwargs) -> Any:
        """
        Create a new entity of the specified type.

        Args:
            entity_type: The type of entity to create (study, run, logical_worker, physical_worker).

        Returns:
            The created entity.

        Raises:
            EntityManagerNotSupportedError: If the entity type is not supported.
        """
        self._validate_entity_type(entity_type)
        return self._entity_managers[entity_type].create(*args, **kwargs)

    def get(self, entity_type: str, *args, **kwargs) -> Any:
        """
        Get an entity by type and identifier.

        Args:
            entity_type: The type of entity to get (study, run, logical_worker, physical_worker).

        Returns:
            The requested entity.

        Raises:
            EntityManagerNotSupportedError: If the entity type is not supported.
        """
        self._validate_entity_type(entity_type)
        return self._entity_managers[entity_type].get(*args, **kwargs)

    def get_all(self, entity_type: str) -> List[Any]:
        """
        Get all entities of a specific type.

        Args:
            entity_type: The type of entities to get (study, run, logical_worker, physical_worker).

        Returns:
            A list of all entities of the specified type.

        Raises:
            EntityManagerNotSupportedError: If the entity type is not supported.
        """
        self._validate_entity_type(entity_type)
        return self._entity_managers[entity_type].get_all()

    def delete(self, entity_type: str, *args, **kwargs) -> None:
        """
        Delete an entity by type and identifier.

        Args:
            entity_type: The type of entity to delete (study, run, logical_worker, physical_worker).

        Raises:
            EntityManagerNotSupportedError: If the entity type is not supported.
        """
        self._validate_entity_type(entity_type)
        self._entity_managers[entity_type].delete(*args, **kwargs)

    def delete_all(self, entity_type: str, **kwargs) -> None:
        """
        Delete all entities of a specific type.

        Args:
            entity_type: The type of entities to delete (study, run, logical_worker, physical_worker).

        Raises:
            EntityManagerNotSupportedError: If the entity type is not supported.
        """
        self._validate_entity_type(entity_type)
        self._entity_managers[entity_type].delete_all(**kwargs)

    def get_everything(self) -> List[Any]:
        """
        Get all entities from all entity managers.

        Returns:
            A dictionary mapping entity types to lists of entities.
        """
        result = []
        for manager in self._entity_managers.values():
            result.extend(manager.get_all())
        return result

    def delete_everything(self, force: bool = False) -> None:
        """
        Delete all entities from all entity managers.

        This method deletes studies last to ensure proper cleanup of dependencies.
        """
        flush_database = False
        if force:
            flush_database = True
        else:
            # Ask the user for confirmation
            valid_inputs = ["y", "n"]
            user_input = input("Are you sure you want to flush the entire database? (y/n): ").strip().lower()
            while user_input not in valid_inputs:
                user_input = input("Invalid input. Use 'y' for 'yes' or 'n' for 'no': ").strip().lower()

            if user_input == "y":
                flush_database = True

        if flush_database:
            LOG.info("Flushing the database...")
            self.backend.flush_database()
            LOG.info("Database successfully flushed.")
        else:
            LOG.info("Database flush cancelled.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""This module represents all of the logic that goes into a step"""

import logging
import os
import re
from contextlib import suppress
from copy import deepcopy
from typing import Dict, List, Tuple

from celery import current_task
from maestrowf.abstracts.enums import State
from maestrowf.abstracts.interfaces.scriptadapter import ScriptAdapter
from maestrowf.datastructures.core.executiongraph import _StepRecord
from maestrowf.datastructures.core.study import StudyStep
from maestrowf.interfaces.script import SubmissionRecord

from merlin.common.enums import ReturnCode
from merlin.study.script_adapter import MerlinScriptAdapter
from merlin.study.status import read_status, write_status
from merlin.utils import needs_merlin_expansion


LOG = logging.getLogger(__name__)


def get_current_worker() -> str:
    """
    Get the worker on the current running task from Celery.

    This function retrieves the name of the worker that is currently
    executing the task. It extracts the worker's name from the task's
    request hostname.

    Returns:
        The name of the current worker.
    """
    worker = re.search(r"@.+\.", current_task.request.hostname).group()
    worker = worker[1 : len(worker) - 1]
    return worker


def get_current_queue() -> str:
    """
    Get the queue on the current running task from Celery.

    This function retrieves the name of the queue that the current
    task is associated with. It extracts the routing key from the
    task's delivery information and removes the queue tag defined
    in the configuration.

    Returns:
        The name of the current queue.
    """
    from merlin.config.configfile import CONFIG  # pylint: disable=C0415

    queue = current_task.request.delivery_info["routing_key"]
    queue = queue.replace(CONFIG.celery.queue_tag, "")
    return queue


class MerlinStepRecord(_StepRecord):
    """
    This class is a wrapper for the Maestro `_StepRecord` to remove
    a re-submit message and handle status updates.

    Attributes:
        condensed_workspace (str): A condensed version of the workspace path.
        elapsed_time (str): The total elapsed time for the step execution.
        jobid (List[int]): A list of job identifiers assigned by the scheduler.
        maestro_step (StudyStep): The StudyStep object associated with this step.
        merlin_step (Step): The Step object associated with this step.
        restart_limit (int): Upper limit on the number of restart attempts.
        restart_script (str): Script to resume record execution (if applicable).
        run_time (str): The run time for the step execution.
        status (State): The current status of the step.
        to_be_scheduled (bool): Indicates if the record needs scheduling.
        workspace (Variable): The output workspace for this step, represented as a Variable.

    Methods:
        mark_end: Marks the end of the step with the given state.
        mark_restart: Increments the restart count for the step.
        mark_running: Marks the step as running and updates the status file.
        setup_workspace: Initializes the workspace and status file for the step.
    """

    def __init__(self, workspace: str, maestro_step: StudyStep, merlin_step: "Step", **kwargs):
        """
        Initializes the `MerlinStepRecord` class which helps track the status of a step.

        Args:
            workspace: The output workspace for this step.
            maestro_step: The
                [StudyStep](https://maestrowf.readthedocs.io/en/latest/Maestro/reference_guide/api_reference/datastructures/core/index.html#maestrowf.datastructures.core.StudyStep)
                object associated with this step.
            merlin_step: The [Step][study.step.Step] object associated with this step.
        """
        _StepRecord.__init__(self, workspace, maestro_step, status=State.INITIALIZED, **kwargs)
        self.merlin_step: Step = merlin_step

    @property
    def condensed_workspace(self) -> str:
        """
        Generate a condensed version of the workspace path for display purposes.

        This property constructs a shorter representation of the workspace path by extracting relevant
        components based on the study name and a timestamp pattern. If a match is found using a regular
        expression, the workspace path is split to isolate the condensed portion. If no match is found,
        a fallback method is used to manually create a condensed path based on the step name.

        Returns:
            A string representing the condensed workspace path, which is easier to read and display.
        """
        timestamp_regex = r"\d{8}-\d{6}/"
        match = re.search(rf"{self.merlin_step.study_name}_{timestamp_regex}", self.workspace.value)

        # If we got a match from the regex (which we should always get) then use it to condense the workspace
        if match:
            condensed_workspace = self.workspace.value.split(match.group())[1]
        # Otherwise manually condense (which could have issues if step names/parameters/study names are equivalent)
        else:
            step_name = self.merlin_step.name_no_params()
            end_of_path = self.workspace.value.rsplit(step_name, 1)[1]
            condensed_workspace = f"{step_name}{end_of_path}"

        LOG.debug(f"Condense workspace '{condensed_workspace}'")
        return condensed_workspace

    def _execute(self, adapter: ScriptAdapter, script: str) -> Tuple[SubmissionRecord, int]:
        """
        Executes the script using the provided adapter, overriding the default behavior to ensure
        that the step is marked as running and to facilitate job submission.

        This method overrides the `_execute` method from the base class `_StepRecord` in Maestro.
        It ensures that `self.to_be_scheduled` is always true, allowing for the invocation of
        `self.mark_running()` to update the status of the step.

        Args:
            adapter: The script adapter used to submit jobs.
            script: The script to be submitted to the script adapter.

        Returns:
            A tuple containing the return code and the job identifier from the execution of the script.
        """
        self.mark_running()

        LOG.info(f"Submitting script for {self.name}")
        srecord = adapter.submit(self.step, script, self.workspace.value)

        retcode = srecord.submission_code
        jobid = srecord.job_identifier
        return retcode, jobid

    def mark_running(self):
        """Mark the start time of the record and update the status file."""
        LOG.debug(f"Marking running for {self.name}")
        super().mark_running()
        self._update_status_file()

    def mark_end(self, state: ReturnCode, max_retries: bool = False):
        """
        Marks the end time of the record with the associated termination state
        and updates the status file.

        This method logs the action of marking the end of the step, maps the provided
        termination state to a corresponding Maestro state and result, and updates
        the status file accordingly. If the maximum number of retries has been reached
        for a soft failure, it appends a message to the result.

        Args:
            state: A ReturnCode object representing the end state of the task.
            max_retries: A flag indicating whether the maximum number of retries has been reached.
        """
        LOG.debug(f"Marking end for {self.name}")

        # Dictionary to keep track of associated variables for each return code
        state_mapper = {
            ReturnCode.OK: {
                "maestro state": State.FINISHED,
                "result": "MERLIN_SUCCESS",
            },
            ReturnCode.DRY_OK: {
                "maestro state": State.DRYRUN,
                "result": "MERLIN_DRY_SUCCESS",
            },
            ReturnCode.RETRY: {
                "maestro state": State.FINISHED,
                "result": "MERLIN_RETRY",
            },
            ReturnCode.RESTART: {
                "maestro state": State.FINISHED,
                "result": "MERLIN_RESTART",
            },
            ReturnCode.SOFT_FAIL: {
                "maestro state": State.FAILED,
                "result": "MERLIN_SOFT_FAIL",
            },
            ReturnCode.HARD_FAIL: {
                "maestro state": State.FAILED,
                "result": "MERLIN_HARD_FAIL",
            },
            ReturnCode.STOP_WORKERS: {
                "maestro state": State.CANCELLED,
                "result": "MERLIN_STOP_WORKERS",
            },
            "UNKNOWN": {
                "maestro state": State.UNKNOWN,
                "result": "MERLIN_UNRECOGNIZED",
            },
        }

        # Check if the state provided is valid
        if state not in state_mapper:
            state = "UNKNOWN"

        # Call to super().mark_end() will mark end time and update self.status for us
        super().mark_end(state_mapper[state]["maestro state"])
        step_result = state_mapper[state]["result"]

        # Append a "max retries reached" message to the step result if necessary
        if state == ReturnCode.SOFT_FAIL and max_retries:
            step_result += " (MAX RETRIES REACHED)"

        # Update the status file
        self._update_status_file(result=step_result)

    def mark_restart(self):
        """Increment the number of restarts we've had for this step and update the status file."""
        LOG.debug(f"Marking restart for {self.name}")
        if self.restart_limit == 0 or self._num_restarts < self.restart_limit:
            self._num_restarts += 1
            self._update_status_file()

    def setup_workspace(self):
        """Initialize the record's workspace and status file."""
        LOG.debug(f"Setting up workspace for {self.name}")
        super().setup_workspace()
        self._update_status_file()

    def _update_status_file(
        self,
        result: str = None,
        task_server: str = "celery",
    ):
        """
        Constructs a dictionary containing status information and creates a signature
        for the update_status Celery task. This signature is executed within the method.

        This method checks if a status file already exists; if it does, it updates the
        existing file with the current status information. If not, it initializes a new
        status dictionary. The method also includes optional parameters for the result
        of the task and the task server being used.

        Args:
            result: An optional string representation of a ReturnCode value, applied
                when the step has finished running.
            task_server: An optional parameter to specify the task server being used.
        """

        # This dict is used for converting an enum value to a string for readability
        state_translator: Dict[State, str] = {
            State.INITIALIZED: "INITIALIZED",
            State.RUNNING: "RUNNING",
            State.FINISHED: "FINISHED",
            State.CANCELLED: "CANCELLED",
            State.DRYRUN: "DRY_RUN",
            State.FAILED: "FAILED",
            State.UNKNOWN: "UNKNOWN",
        }

        LOG.debug(f"Marking status for {self.name} as {state_translator[self.status]}.")
        if result:
            LOG.debug(f"Result for {self.name} is {result}")

        status_filepath = f"{self.workspace.value}/MERLIN_STATUS.json"

        LOG.debug(f"Status filepath for {self.name}: '{status_filepath}")

        # If the status file already exists then we can just add to it
        if os.path.exists(status_filepath):
            status_info = read_status(status_filepath, f"{self.workspace.value}/status.lock")
        else:
            # Create the parameter entries
            cmd_params = restart_params = None
            if self.merlin_step.params["cmd"]:
                cmd_params = dict(self.merlin_step.params["cmd"].items())
            if self.merlin_step.params["restart_cmd"]:
                restart_params = dict(self.merlin_step.params["restart_cmd"].items())
            LOG.debug(f"Cmd parameters for {self.name}: {cmd_params}; Restart params: {restart_params}")

            # Inititalize the status_info dict we'll be dumping to the status file
            status_info = {
                self.name: {
                    "parameters": {
                        "cmd": cmd_params,
                        "restart": restart_params,
                    }
                }
            }

        # Put together a dict of status info
        status_info[self.name][self.condensed_workspace] = {
            "status": state_translator[self.status],
            "return_code": result,
            "elapsed_time": self.elapsed_time,
            "run_time": self.run_time,
            "restarts": self.restarts,
        }

        # Add celery specific info
        if task_server == "celery":
            from merlin.celery import app  # pylint: disable=C0415

            # If the tasks are always eager, this is a local run and we won't have workers running
            if not app.conf.task_always_eager:
                status_info[self.name]["task_queue"] = get_current_queue()

                # Add the current worker to the workspace-specific status info
                current_worker = get_current_worker()
                if "workers" not in status_info[self.name][self.condensed_workspace]:
                    status_info[self.name][self.condensed_workspace]["workers"] = [current_worker]
                elif current_worker not in status_info[self.name][self.condensed_workspace]["workers"]:
                    status_info[self.name][self.condensed_workspace]["workers"].append(current_worker)

                # Add the current worker to the overall-step status info
                if "workers" not in status_info[self.name]:
                    status_info[self.name]["workers"] = [current_worker]
                elif current_worker not in status_info[self.name]["workers"]:
                    status_info[self.name]["workers"].append(current_worker)

        LOG.info(f"Writing status for {self.name} to '{status_filepath}...")
        write_status(status_info, status_filepath, f"{self.workspace.value}/status.lock")
        LOG.info(f"Status for {self.name} successfully written.")


class Step:
    """
    This class provides an abstraction for an execution step, which can be
    executed by calling the [`execute`][study.step.Step.execute] method.

    Attributes:
        max_retries (int): Returns the maximum number of retries for this step.
        mstep (_StepRecord): The Maestro StepRecord object associated with this step.
        parameter_info (dict): A dictionary containing information about parameters in the study.
        params (Dict): A dictionary containing command parameters for the step, including 'cmd' and 'restart_cmd'.
        restart (bool): Property to get or set the restart status of the step.
        retry_delay (int): Returns the retry delay for the step (default is 1).
        study_name (str): The name of the study this step belongs to.

    Methods:
        check_if_expansion_needed: Checks if command expansion is needed based on specified labels.
        clone_changing_workspace_and_cmd: Produces a deep copy of the current step, with optional command
            and workspace modifications.
        establish_params: Pulls parameters from the step parameter map if applicable.
        execute: Executes the step using the provided adapter configuration.
        get_cmd: Retrieves the run command text body.
        get_restart_cmd: Retrieves the restart command text body, or None if not available.
        get_task_queue: Retrieves the task queue for the step.
        get_task_queue_from_dict: Static method to get the task queue from a step dictionary.
        get_workspace: Retrieves the workspace where this step is to be executed.
        name: Retrieves the name of the step.
        name_no_params: Gets the original name of the step without parameters or sample labels.
    """

    def __init__(self, maestro_step_record: _StepRecord, study_name: str, parameter_info: Dict):
        """
        Initializes the `Step` object which acts as a way to track everything about a step.

        Args:
            maestro_step_record: The `StepRecord` object.
            study_name: The name of the study
            parameter_info: A dict containing information about parameters in the study
        """
        self.mstep: _StepRecord = maestro_step_record
        self.study_name: str = study_name
        self.parameter_info: Dict = parameter_info
        self.__restart: bool = False
        self.params: Dict = {"cmd": {}, "restart_cmd": {}}
        self.establish_params()

    def get_cmd(self) -> str:
        """
        Retrieve the run command text body for the step.

        Returns:
            The run command text body for the step.
        """
        return self.mstep.step.__dict__["run"]["cmd"]

    def get_restart_cmd(self) -> str:
        """
        Retrieve the restart command text body for the step.

        Returns:
            The restart command text body for the step, or None if no restart command is available.
        """
        return self.mstep.step.__dict__["run"]["restart"]

    def clone_changing_workspace_and_cmd(
        self,
        new_cmd: str = None,
        cmd_replacement_pairs: List[Tuple[str]] = None,
        new_workspace: str = None,
    ) -> "Step":
        """
        Produces a deep copy of the current step, with optional modifications to
        the command and workspace, performing variable substitutions as we go.

        This method creates a new instance of the Step class by cloning the
        current step and allowing for modifications to the command text and
        workspace. It performs variable substitutions in the command based on
        the provided replacement pairs.

        Args:
            new_cmd: If provided, replaces the existing command with this new command.
            cmd_replacement_pairs: A list of pairs where each pair contains a string to
                be replaced and its replacement. The method will perform replacements in
                both the run command and the restart command.
            new_workspace: If provided, sets this as the workspace for the new step. If
                not specified, the current workspace will be used.

        Returns:
            A new Step instance with the modified command and workspace.
        """
        LOG.debug(f"clone called with new_workspace {new_workspace}")
        step_dict = deepcopy(self.mstep.step.__dict__)

        if new_cmd is not None:
            step_dict["run"]["cmd"] = new_cmd

        if cmd_replacement_pairs is not None:
            for str1, str2 in cmd_replacement_pairs:
                cmd = step_dict["run"]["cmd"]
                step_dict["run"]["cmd"] = re.sub(re.escape(str1), str2, cmd, flags=re.I)

                restart_cmd = step_dict["run"]["restart"]
                if restart_cmd:
                    step_dict["run"]["restart"] = re.sub(re.escape(str1), str2, restart_cmd, flags=re.I)

        if new_workspace is None:
            new_workspace = self.get_workspace()
        LOG.debug(f"cloned step with workspace {new_workspace}")
        study_step = StudyStep()
        study_step.name = step_dict["_name"]
        study_step.description = step_dict["description"]
        study_step.run = step_dict["run"]
        return Step(MerlinStepRecord(new_workspace, study_step, self), self.study_name, self.parameter_info)

    def get_task_queue(self) -> str:
        """
        Retrieve the task queue for the current Step.

        Returns:
            The name of the task queue for the Step, which may be influenced
                by the configuration settings.
        """
        return self.get_task_queue_from_dict(self.mstep.step.__dict__)

    @staticmethod
    def get_task_queue_from_dict(step_dict: Dict) -> str:
        """
        Get the task queue from a given Maestro step dictionary.

        This static method extracts the task queue information from the
        provided step dictionary. It considers the configuration settings
        to determine the appropriate queue name, including handling cases
        where the task queue may be omitted.

        Args:
            step_dict: A dictionary representation of a Maestro step, expected
                to contain a "run" key with a "task_queue" entry.

        Returns:
            The name of the task queue. If the task queue is not specified
                or is set to "none", it returns the default queue name based
                on the configuration.
        """
        from merlin.config.configfile import CONFIG  # pylint: disable=C0415

        queue_tag = CONFIG.celery.queue_tag
        omit_tag = CONFIG.celery.omit_queue_tag
        if omit_tag:
            queue = "merlin"
        else:
            queue = queue_tag

        with suppress(TypeError, KeyError):
            val = step_dict["run"]["task_queue"]
            if not (val is None or val.lower() == "none" or val == ""):
                if omit_tag:
                    queue = val
                else:
                    queue = queue_tag + val
        return queue

    @property
    def retry_delay(self) -> int:
        """
        Get the retry delay for the step.

        Returns:
            The retry delay in seconds. Defaults to 1 if not specified.
        """
        default_retry_delay = 1
        return self.mstep.step.__dict__["run"].get("retry_delay", default_retry_delay)

    @property
    def max_retries(self) -> int:
        """
        Get the maximum number of retries for this step.

        Returns:
            The maximum number of retries for the step.
        """
        return self.mstep.step.__dict__["run"]["max_retries"]

    @property
    def restart(self) -> bool:
        """
        Get the restart property.

        Returns:
            True if the step is set to restart, False otherwise.
        """
        return self.__restart

    @restart.setter
    def restart(self, val: bool):
        """
        Set the restart property.

        Args:
            val: The new value for the restart property. It should be
                a boolean value indicating whether the step should restart.
        """
        self.__restart = val

    def establish_params(self):
        """
        Establish parameters for the step from the parameter map.

        This method checks if the current step uses parameters by accessing
        the `step_param_map` from `parameter_info`. If parameters are found
        for the current step, it updates the `params` dictionary with the
        corresponding values.
        """
        try:
            step_params = self.parameter_info["step_param_map"][self.name()]
            for cmd_type in step_params:
                self.params[cmd_type].update(step_params[cmd_type])
        except KeyError:
            pass

    def check_if_expansion_needed(self, labels: List[str]) -> bool:
        """
        Check if expansion is needed based on commands and labels.

        This method determines whether the command associated with the
        current step requires expansion. It checks for the presence of
        default keywords or specified sample column labels.

        Args:
            labels: A list of labels to check against the commands.

        Returns:
            True if the command requires expansion, False otherwise.
        """
        return needs_merlin_expansion(self.get_cmd(), self.get_restart_cmd(), labels)

    def get_workspace(self) -> str:
        """
        Get the workspace for the current step.

        Returns:
            The workspace associated with this step.
        """
        return self.mstep.workspace.value

    def name(self) -> str:
        """
        Get the name of the current step.

        Returns:
            The name of the step.
        """
        return self.mstep.step.__dict__["_name"]

    def name_no_params(self) -> str:
        """
        Get the original name of the step without parameters or sample labels.

        This method retrieves the name of the step and removes any
        parameter labels or sample identifiers that may be included
        in the name. It ensures that the returned name is clean and
        free from extraneous characters, such as trailing periods or
        underscores.

        Returns:
            The cleaned name of the step, free from parameters and sample labels.
        """
        # Get the name with everything still in it
        name = self.name()

        # Remove the parameter labels from the name
        for label in self.parameter_info["labels"]:
            name = name.replace(f"{label}", "")

        # Remove possible leftover characters after condensing the name
        while name.endswith(".") or name.endswith("_"):
            if name.endswith("."):
                split_char = "."
            else:
                split_char = "_"
            split_name = name.rsplit(split_char, 1)
            name = "".join(split_name)

        return name

    def execute(self, adapter_config: Dict) -> ReturnCode:
        """
        Execute the step with the provided adapter configuration.

        This method performs the execution of the step by configuring
        the necessary parameters and invoking the appropriate adapter.
        It updates the adapter configuration based on the step's
        requirements, sets up the workspace, and generates the script
        for execution. If a dry run is specified, it prepares the
        workspace without executing any tasks.

        Args:
            adapter_config (dict): A dictionary containing configuration
                for the maestro script adapter, including:\n
                - `shell`: The shell to use for execution.
                - `batch_type`: The type of batch processing to use.
                - `dry_run`: A boolean indicating whether to perform a
                dry run (setup only, no execution).

        Returns:
            (common.enums.ReturnCode): A [`ReturnCode`][common.enums.ReturnCode] object representing
                the result of the execution.
        """
        # Update shell if the task overrides the default value from the batch section
        default_shell = adapter_config.get("shell")
        shell = self.mstep.step.run.get("shell", default_shell)
        adapter_config.update({"shell": shell})

        # Update batch type if the task overrides the default value from the batch section
        default_batch_type = adapter_config.get("batch_type", adapter_config["type"])
        # Set batch_type to default if unset
        adapter_config.setdefault("batch_type", default_batch_type)
        # Override the default batch: type: from the step config
        batch = self.mstep.step.run.get("batch", None)
        if batch:
            batch_type = batch.get("type", default_batch_type)
            adapter_config.update({"batch_type": batch_type})

        adapter = MerlinScriptAdapter(**adapter_config)
        LOG.debug(f"Maestro step config = {adapter_config}")

        # Preserve the default shell if the step shell is different
        adapter_config.update({"shell": default_shell})
        # Preserve the default batch type if the step batch type is different
        adapter_config.update({"batch_type": default_batch_type})

        self.mstep.setup_workspace()
        self.mstep.generate_script(adapter)

        # dry run: sets up a workspace without executing any tasks. Each step's
        # workspace directory is created, and each step's command script is
        # written to it. The command script is not run, so there is no
        # 'MERLIN_FINISHED' file, nor '<step>.out' nor '<step>.err' log files.
        if adapter_config["dry_run"] is True:
            return ReturnCode.DRY_OK

        # TODO: once maestrowf is updated so that execute returns a
        # submissionrecord, then we need to return the record.return_code here
        # at that point, we can drop the use of MerlinScriptAdapter above, and
        # go back to using the adapter specified by the adapter_config['type']
        # above
        # If the above is done, then merlin_step in tasks.py can be changed to
        # calls to the step execute and restart functions.
        if self.restart and self.get_restart_cmd():
            return ReturnCode(self.mstep.restart(adapter))

        return ReturnCode(self.mstep.execute(adapter))
